# --- built in ---
import os
import sys
import time
import logging

from typing import Union

# --- 3rd party ---
import gym
import numpy as np

# --- my module ---
from unstable_baselines.lib import utils as ub_utils
from unstable_baselines.lib.envs.vec import base as vec_base

__all__ = [
    'DummyVecEnv',
    'VecEnv'
]

class EnvWorker(vec_base.BaseEnvWorker):
    def __init__(self, env_fn, auto_reset: bool):
        self.env = env_fn()
        self._res = None
        super().__init__(env_fn, auto_reset)
    
    def getattr(self, attrname: str):
        return getattr(self.env, attrname)

    def setattr(self, attrname: str, value):
        return setattr(self.env, attrname, value)

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

    def step_async(self, act):
        obs, rew, done, info = self.env.step(act)
        if self._auto_reset and done:
            obs = self.env.reset()
        self._res = (obs, rew, done, info)

    def step_wait(self):
        return self._res

    def seed(self, seed):
        super().seed(seed)
        return self.env.seed(seed)

    def render(self, **kwargs):
        return self.env.render(**kwargs)

    def close_async(self):
        self.env.close()

    def close_wait(self):
        pass


class DummyVecEnv(vec_base.BaseVecEnv):
    def __init__(self,
        env_fns: list, 
        rms_norm: Union[str, bool, ub_utils.RMSNormalizer] = False,
        auto_reset: bool = True,
    ):
        super().__init__(env_fns, EnvWorker, rms_norm, auto_reset)


class VecEnv(vec_base.BaseVecEnv):
    def __init__(self, 
        envs: list,
        rms_norm: Union[str, bool, ub_utils.RMSNormalizer] = False,
        auto_reset: bool = True,
    ):
        env_fns = [lambda i=j: envs[i] for j in range(len(envs))]
        super().__init__(env_fns, EnvWorker, rms_norm, auto_reset)


import re
import os
import sys
import time
import logging

from collections import deque

import subprocess

from .base import AsciiFont

import cloudpickle as pickle

DEFAULT_PKL_PATH = './pkl'
DEFAULT_PKL_PATH = os.path.join(
        os.path.dirname(os.path.abspath(__file__)), 
        DEFAULT_PKL_PATH)
# absolute path

FONT_TYPE_LIST = {}


def load_font_type_list(installed_path=DEFAULT_PKL_PATH):
    
    for pkl in os.listdir(installed_path):
        
        if not pkl.endswith('.pkl'):
            continue

        pkl_path = os.path.join(installed_path, pkl)
        pkl_name, _ = os.path.splitext(pkl)
        for rep in [' ', '/', '\\', '-']:
            pkl_name = pkl_name.replace(rep, '_')

        FONT_TYPE_LIST[pkl_name] = {'path': pkl_path, 'module': None}


        

def install_font_type(name, flf_path):

    flf_name = name
    for rep in [' ', '/', '\\', '-']:
        flf_name = flf_name.replace(rep, '_')

    if flf_name in FONT_TYPE_LIST:
        print('Font type \'{}\' is already installed'.format(flf_name))

    if not os.path.isfile(flf_path):
        raise Exception('Path \'{}\' does not exist'.format(flf_path))

    FONT_TYPE_LIST[flf_name] = {'path':flf_path, 'module': None}

'''
def load(font_type):

    flf_path = FONT_TYPE_LIST[font_type]['path']


    HEADER = 'flf2a'
    
    line_end = ['#', '@', chr(127)]
    font_end = ['##', '@@', chr(127)*2]

    charset_pattern = r'.*charset=(.*)$'
    matcher = re.compile(charset_pattern)

    out = subprocess.check_output('file -i "{}"'.format(flf_path), shell=True).decode('utf-8')

    match = matcher.match(out)

    if match is not None:
        charset = match.group(1).strip()
    else:
        charset = 'utf-8'

    if charset == 'binary':
        print('Detect binary encoding')
        charset = 'iso-8859-1'

    with open(flf_path, 'r', encoding=charset) as inpf:

        header = inpf.readline()
        magic_number = header.split(' ')

        assert magic_number[0].startswith(HEADER), 'Got an unexpected header: {}'.format(magic_number[0])


        hardblank = magic_number[0].replace(HEADER, '')
        ch_height = int(magic_number[1])

        # create read buffer
        read_buf = deque(maxlen=ch_height)

        font = AsciiFont(hardblank)

        current_ascii = 32
        stop_ascii = 127

        for line in inpf:
            line = line.replace('\n', '').rstrip()

            if not line:
                continue

            
            if not (line[-1] in line_end):
                pass
            else:
                # font end
                if not ((line[-2:] in font_end) and (len(read_buf) == read_buf.maxlen-1)):

                    #if (not (line[-2:]) in font_end):

                    end = line[-1]
                    line = line.replace(end, '\n')

                    read_buf.append(line)

                else:

                    end = line[-2:]
                    line = line.replace(end, '')

                    read_buf.append(line)

                    char = chr(current_ascii)

                    # concat to one string ('\n')
                    font[char] = ''.join(read_buf)

                    current_ascii += 1

                    read_buf.clear()

            if current_ascii == stop_ascii:
                break
    
    FONT_TYPE_LIST[font_type]['module'] = font
    
    return font
'''


def load(font_type):
    pkl_path = FONT_TYPE_LIST[font_type]['path']
    
    
    with open(pkl_path, 'rb') as inpf:
    
        font_dict = pickle.load(inpf)
        
    ascii_font = AsciiFont(font_dict['hardblank'])
    for k, v in font_dict.items():
        if k == 'hardblank':
            continue
        ascii_font[k] = v
    
    FONT_TYPE_LIST[font_type]['module'] = ascii_font
    
    return ascii_font
    
    
def retrieves(font_type, chars):
    if font_type in FONT_TYPE_LIST:
        if FONT_TYPE_LIST[font_type]['module'] is not None:
            font_type = FONT_TYPE_LIST[font_type]['module']
        else:
            font_type = load(font_type)
    else:
        raise Exception('Font type \'{}\' is not installed'.format(font_type))

    type_chars = []

    for c in chars:
        type_char = font_type.retrieve(c)

        type_chars.append(type_char)


    return type_chars


load_font_type_list()



import re
import os
import sys
import time
import logging

from collections import deque

import subprocess

import cloudpickle
from base import AsciiFont

TEXT_FLF = 'Rectangles.flf'

HEADER = 'flf2a'

line_end = ['#', '@', chr(127)]
font_end = ['##', '@@', chr(127)*2]

charset_pattern = r'.*charset=(.*)$'
matcher = re.compile(charset_pattern)


def _compile(input_filename, object_name):
    out = subprocess.check_output('file -i "{}"'.format(input_filename), shell=True).decode('utf-8')


    match = matcher.match(out)

    if match is not None:
        charset = match.group(1).strip()
    else:
        charset = 'utf-8'

    if charset == 'binary':
        print('Detect binary encoding')
        charset = 'iso-8859-1'

    print('    Using charactor set: {}'.format(charset))
    with open(input_filename, 'r', encoding=charset) as inpf:
        header = inpf.readline()
        magic_number = header.split(' ')

        assert magic_number[0].startswith(HEADER), 'Got an unexpected header: {}'.format(magic_number[0])

        hardblank = magic_number[0].replace(HEADER, '')
        ch_height = int(magic_number[1])

        # create read buffer
        read_buf = deque(maxlen=ch_height)
        
        font = {'hardblank': hardblank}
        
        current_ascii = 32
        stop_ascii = 127
        
        for line in inpf:
            line = line.replace('\n', '').rstrip()
            
            if not line:
                continue
                
                
            if not (line[-1] in line_end):
                pass
            else:
                # font end
                if not ((line[-2:] in font_end) and (len(read_buf) == read_buf.maxlen-1)):

                    #if (not (line[-2:]) in font_end):

                    end = line[-1]
                    line = line.replace(end, '\n')

                    read_buf.append(line)
                else:

                    end = line[-2:]
                    line = line.replace(end, '')

                    read_buf.append(line)

                    char = chr(current_ascii)

                    # concat to one string ('\n')
                    font[char] = ''.join(read_buf)

                    current_ascii += 1

                    read_buf.clear()
                    
                    
            if current_ascii == stop_ascii:
                break
    
    return font


if __name__ == '__main__':

    assert len(sys.argv) >= 2, 'python compile.py [flf files]'

    for argv in sys.argv[1:]:

        input_filename = str(argv)

        basename = os.path.basename(input_filename)

        object_name, ext = basename.split('.')

        # convert to valid name
        for rep in [' ', '/', '\\', '-']:
            object_name = object_name.replace(rep, '_')
        
        output_filename = os.path.join('pkl', object_name + '.pkl')

        print('Compiling file "{}" to "{}"'.format(input_filename, output_filename))

        try:
            font_dict = _compile(input_filename, object_name)
        except Exception as e:
            print('    Failed to generate file: {}'.format(str(e)))
            continue

        print('    test loadability:')

        try:
            if not os.path.isfile(object_name):
                time.sleep(0.1)
                
            ascii_font = AsciiFont(font_dict['hardblank'])
            for k, v in font_dict.items():
                if k == 'hardblank':
                    continue
                ascii_font[k] = v
                
            print(ascii_font['a'])
            print(ascii_font['A'])
            print(ascii_font['0'])
            print('        Clear')
        except:
            print('        Failed')
            raise
            
        # save ascii_font to pkl file
        with open(output_filename, 'wb') as outf:
            cloudpickle.dump(font_dict, outf)
            
        print('Save to {}'.format(output_filename))

'''
def generate(input_filename, output_filename, object_name):

    out = subprocess.check_output('file -i "{}"'.format(input_filename), shell=True).decode('utf-8')


    match = matcher.match(out)

    if match is not None:
        charset = match.group(1).strip()
    else:
        charset = 'utf-8'

    if charset == 'binary':
        print('Detect binary encoding')
        charset = 'iso-8859-1'

    print('    Using charactor set: {}'.format(charset))
    with open(input_filename, 'r', encoding=charset) as inpf, open(output_filename, 'w') as outf:
        header = inpf.readline()
        magic_number = header.split(' ')

        assert magic_number[0].startswith(HEADER), 'Got an unexpected header: {}'.format(magic_number[0])

        hardblank = magic_number[0].replace(HEADER, '')
        ch_height = int(magic_number[1])

        # create read buffer
        read_buf = deque(maxlen=ch_height)
        
        # create output file header
        outf.write(\
"""

if not '.' in __name__:
    package_name = '__main__'
else:
    package_name = __name__.rsplit(".", 1)[0]

if package_name == "__main__":
    module_name = 'base'
else:
    module_name = package_name + '.' + 'base'

module = __import__(module_name, fromlist=['.'])
AsciiFont = getattr(module, 'AsciiFont')


"""     )
        
        outf.write('{} = AsciiFont("{}")\n'.format(object_name, hardblank))
        
        outf.write('# {}\n\n'.format(header))

        # start from 32 (space)
        current_ascii = 32
        stop_ascii = 127

        for line in inpf:
            # remove newlines
            line = line.replace('\n', '').rstrip()

            #print('line: {}/line end: {}'.format(line, line[-1]))

            if not line:
                continue

            # write comments
            if not (line[-1] in LINE_END):
                outf.write('# {}\n'.format(line))
            else:

                # replace hardblank to normal space
                #line = line.replace(hardblank, ' ')
                
                # if not font end
                if not (line[-2:] in FONT_END):

                    # replace line end to new line
                    for line_end in LINE_END:
                        line = line.replace(line_end, '\n')

                    # append to read_buf
                    read_buf.append(line)

                # font end
                else:

                    # remove font end
                    for font_end in FONT_END:
                        line = line.replace(font_end, '')

                    # append to buf
                    read_buf.append(line)

                    char = chr(current_ascii)

                    if char == '"':
                        char = '\\"'
                    elif char == '\\':
                        char = '\\\\'


                    outf.write('{}["{}"] = (\n'.format(object_name, char))
                    for buf in read_buf:
                        # convert all escape to literal
                        buf = buf.replace('\\', '\\\\').replace('"', '\\"').replace('\n', '\\n')
                        outf.write('"{}"\n'.format(buf))

                    outf.write(')\n\n')

                    current_ascii += 1
                    
            if current_ascii == stop_ascii:
                break

    return True


if __name__ == '__main__':

    assert len(sys.argv) >= 2, 'python generator.py [flf files]'

    for argv in sys.argv[1:]:

        input_filename = str(argv)

        basename = os.path.basename(input_filename)

        object_name, ext = basename.split('.')

        # convert to valid name
        for rep in [' ', '/', '\\', '-']:
            object_name = object_name.replace(rep, '_')
        
        output_filename = object_name + '.py'

        print('Generating file "{}" from "{}"'.format(output_filename, input_filename))

        try:
            generate(input_filename, output_filename, object_name)
        except Exception as e:
            print('    Failed to generate file: {}'.format(str(e)))
            continue

        print('    test loadability:')

        try:
            if not os.path.isfile(object_name):
                time.sleep(0.1)

            module = __import__(object_name, fromlist=['.'])
            obj = getattr(module, object_name)
            print('        Clear')
        except:
            print('        Failed')
            raise
'''




class AsciiFont(dict):
    def __init__(self, hardblank='$'):
        self.hardblank = hardblank

    def set_hardblank(self, hardblank):
        self.hardblank = hardblank

    def __getitem__(self, key):
        item = super(AsciiFont, self).__getitem__(key)

        return item.replace(self.hardblank, ' ')


    def retrieve(self, font, remove_hardblank=True):

        item = self.get(font, '')
        #item = super(AsciiFont, self).__getitem__(font)


        if remove_hardblank:
            return item.replace(self.hardblank, ' ')
        else:
            return item



# --- built in ---
import os
import sys
import time
import logging
import datetime

# --- 3rd party ---
import gym
import numpy as np
import tensorflow as tf

# --- my module ---
import unstable_baselines as ub

# === Networks ===

class DiagGaussianPolicyNet(ub.nets.DiagGaussianPolicyNet):
    '''Override Diagonal Gaussian mixture policy net
    with state-independent covariance
    '''
    support_spaces = [gym.spaces.Box]
    def create_logstd_model(self):
        '''(Override) Logstd model'''
        return ub.nets.Constant(tf.Variable(
            np.zeros((self.action_dims,), dtype=np.float32),
            name = 'logstd'
        ))

class PolicyNet(ub.nets.PolicyNet):
    '''Policy network'''
    def __init__(self, action_space,
                       net: tf.keras.Model = None,
                       squash = False, # disabled
                       **kwargs):
        super().__init__(action_space, squash=False, net=net, **kwargs)

    def call(self, inputs, training=True):
        '''Forward policy net

        Args:
            inputs (tf.Tensor): Input tensors, shape (b, *)
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            Distribution: Action distributions.
        '''
        return super().call(inputs, training=training)

    def create_gaussian_policy(self):
        '''(Override) Policy for Box action space'''
        return DiagGaussianPolicyNet(self.action_space, squash=False)

class ValueNet(ub.nets.ValueNet):
    '''Value net'''
    def __init__(self, net: tf.keras.Model=None, **kwargs):
        super().__init__(net=net, **kwargs)

    def call(self, inputs, training=True):
        '''Forward value net

        Args:
            inputs (tf.Tensor): Input tensors, shape (b, *).
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: value predictions, (Default) shape (b, 1)
        '''
        return super().call(inputs, training=training)

    def create_value_model(self):
        '''(Override) Value model arch'''
        return tf.keras.layers.Dense(1)

# === Agent ===

class Agent(ub.base.BaseAgent):
    support_obs_spaces = ub.utils.spaces.All
    support_act_spaces = [gym.spaces.Box, gym.spaces.Discrete]
    def __init__(self, observation_space, 
                       action_space, 
                       share_net: bool = False,
                       force_mlp: bool = False,
                       mlp_units: list = [64, 64],
                       **kwargs):
        '''PPO Agent

        The `observation_space` and `action_space` can be `None` for delayed
        model setup. You should call `set_spaces()` then call `setup()` to 
        manually setup the model.

        Args:
            observation_space (gym.Space): The observation space of the env.
                Set to None for delayed setup.
            action_space (gym.Space): The action space of the env. Set to None 
                for delayed setup.
            share_net (bool, optional): Whether to share the feature extractor 
                between policy and value network. Defaults to False.
            force_mlp (bool, optional): Force to use mlp as feature extractors.
                Default to False.
            mlp_units (list, optional): Number of units for each mlp layer. 
                Defaults to [64, 64]
        '''
        super().__init__(observation_space, action_space, **kwargs)

        self.share_net = share_net
        self.force_mlp = force_mlp
        self.mlp_units = mlp_units

        self.policy = None
        self.value  = None

        if (observation_space is not None) and (action_space is not None):
            self.setup()

    # --- setup group ---
    def _setup_model(self):
        '''Setup models
        Override this method to customize nets
        '''
        # AwesomeNet automatically handles complex observations
        # e.g. Dict or Tuple and creates either NatureCnn or 
        # MlpNet for each space
        create_net_fn = lambda: ub.nets.AwesomeNet(
            input_space = self.observation_space,
            force_mlp   = self.force_mlp,
            mlp_units   = self.mlp_units
        )
        # create base nets, feature extractors
        if self.share_net:
            # share base net
            base_net_p = create_net_fn()
            base_net_v = base_net_p
        else:
            # create separate nets
            base_net_p = create_net_fn()
            base_net_v = create_net_fn()
        # create policy, value nets
        self.policy = PolicyNet(self.action_space, base_net_p)
        self.value  = ValueNet(base_net_v)

    def setup(self):
        '''(Override) Setup agent model'''
        self._setup_model()
        # construct network params
        inputs = ub.utils.input_tensor(self.observation_space)
        self.call_policy(inputs, proc_obs=True)
        self.call_value(inputs, proc_obs=True)

    # --- forwrd group ---
    def call_policy(self, inputs, proc_obs=True, training=True):
        '''Forward policy net'''
        if proc_obs:
            # cast and normalize inputs (eg. image in uint8)
            inputs = self.proc_observation(inputs)
        return self.policy(inputs, training=training)

    def call_value(self, inputs, proc_obs=True, training=True):
        '''Forward value net'''
        if proc_obs:
            # cast and normalize inputs (eg. image in uint8)
            inputs = self.proc_observation(inputs)
        outputs = self.value(inputs, training=training)
        return tf.math.reduce_mean(outputs, axis=-1)

    @tf.function
    def call(self, inputs,
                   proc_obs: bool = True,
                   proc_act: bool = False,
                   det:      bool = False,
                   training: bool = True):
        '''(Override) Forward agent

        Args:
            inputs (tf.Tensor): Batch observations in shape 
                (b, *obs_space.shape), tf.float32
            proc_obs (bool, optional): Preprocess observations. Default to True.
            proc_act (bool, optional): Postprocess actions. Default to False.
            det (bool, optional): Deterministic actions. Defaults to False.
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: Predicted actions, shape (b, *act_space.shape)
            tf.Tensor: Predicted values, shape (b,)
            tf.Tensor: Action log likelihoods, shape (b,)
        '''
        # forward actor
        distrib = self.call_policy(inputs, proc_obs=proc_obs, training=training)
        values  = self.call_value(inputs, proc_obs=proc_obs, training=training)
        # predict actions
        if det:
            actions = distrib.mode()
        else:
            actions = distrib.sample()
        # get log probabilities
        log_probs = distrib.log_prob(actions)
        if proc_act:
            # map actions to env's action space range
            actions = self.proc_action(actions)
        # return results
        return actions, values, log_probs
    
    def predict(self, inputs,
                      proc_obs: bool = True,
                      proc_act: bool = True,
                      det:      bool = True):
        return super().predict(
            inputs, 
            proc_obs=proc_obs,
            proc_act=proc_act,
            det=det
        )
    
    def get_config(self):
        config = super().get_config()
        config.update({
            'share_net':         self.share_net,
            'force_mlp':         self.force_mlp,
            'mlp_units':         self.mlp_units,
        })
        return config

class PPO(ub.base.OnPolicyModel):
    support_obs_spaces = ub.utils.spaces.All
    support_act_spaces = [gym.spaces.Box, gym.spaces.Discrete]
    def __init__(self,
        env,
    # --- hyper parameters ---
        learning_rate: float = 3e-4,
        gamma:         float = 0.99,
        gae_lambda:    float = 0.95,
        policy_clip:   float = 0.05,
        value_clip:    float = None,
        dual_clip:     float = None,
        ent_coef:      float = 0.01,
        vf_coef:       float = 0.5,
        reg_coef:      float = 0.0,
        clipnorm:      float = None,
        target_kl:     float = None,
    # --- architecture parameters ---
        share_net:      bool = False,
        force_mlp:      bool = False,
        mlp_units:      list = [64, 64],
    # --- other parameters ---
        n_steps:         int = 256,
        n_subepochs:     int = 4,
        batch_size:      int = 128,
        **kwargs
    ):
        '''Proximal policy optimization

        The implementation mainly follows its originated paper
        "Proximal Policy Optimization Algorithms" by Schulman et al.

        The `env` can be `None` for delayed model setup. You should call 
        `set_env()` then call `setup()` to manually setup the model.

        Args:
            env (ub.envs.BaseVecEnv): Training env. Must be a vectorized env.
                Set to `None` for delayed model setup.
            learning_rate (float, optional): Learning rate schedule. Can be a
                float or a ub.sche.Scheduler. Defaults to 3e-4.
            gamma (float, optional): Discount factor. Defaults to 0.99.
            gae_lambda (float, optional): GAE smooth parameter. Defaults to 
                0.95.
            policy_clip (float, optional): Policy ratio clip. Defaults to 0.2.
            value_clip (float, optional): Target value clip. Defaults to None.
            dual_clip (float, optional): Policy dual clip from paper "Mastering
                Complex Control in MOBA Games with Deep Reinforcement Learning".
                Defaults to None.
            ent_coef (float, optional): Entropy loss coefficient. Defaults to 
                0.01.
            vf_coef (float, optional): Value loss coefficient. Defaults to 0.5.
            reg_coef (float, optional): Regularization coefficient. Default to 
                0.0.
            clipnorm (float, optional): Gradient clip. Defaults to 0.5.
            target_kl (float, optional): Target KL to early stop. Defaults 
                to None.
            share_net (bool, optional): Whether to share the feature extractor 
                between policy and value network. Defaults to False.
            force_mlp (bool, optional): Force to use mlp as feature extractors.
                Default to False.
            mlp_units (list, optional): Number of units for each mlp layer. 
                Defaults to [64, 64]
            n_steps (int, optional): Number of steps to collect rollouts in 
                each epoch. Defaults to 256.
            n_subepochs (int, optional): Number of subepochs per epoch. Defaults 
                to 4.
            batch_size (int, optional): Training batch size. Defaults to 128.
        '''
        super().__init__(
            env         = env,
            n_steps     = n_steps,
            n_subepochs = n_subepochs,
            batch_size  = batch_size,
            **kwargs
        )
        self.learning_rate = learning_rate
        self.gamma         = gamma
        self.gae_lambda    = gae_lambda
        self.policy_clip   = policy_clip
        self.value_clip    = value_clip
        self.dual_clip     = dual_clip
        self.ent_coef      = ent_coef
        self.vf_coef       = vf_coef
        self.reg_coef      = reg_coef
        self.clipnorm      = clipnorm
        self.target_kl     = target_kl
        self.share_net     = share_net
        self.force_mlp     = force_mlp
        self.mlp_units     = mlp_units
        # initialize objects
        self.buffer    = None
        self.agent     = None
        self.optimizer = None

        if env is not None:
            self.setup()
    
    # --- setup group ---
    def _setup_buffer(self):
        '''Setup buffers
        Override this method to customize to your replay byffer
        '''
        self.buffer = ub.data.DynamicBuffer()
        self.sampler = ub.data.PermuteSampler(self.buffer)
        
    def _setup_model(self):
        '''Setup models, agents
        Override this method to customize to you model
        '''
        self.agent = Agent(
            self.observation_space,
            self.action_space,
            share_net = self.share_net,
            force_mlp = self.force_mlp,
            mlp_units = self.mlp_units
        )

    def _setup_optimizer(self):
        '''Setup optimizers
        Override this method to customize to your optimizer
        '''
        # --- setup scheduler ---
        self.learning_rate = ub.sche.get_scheduler(
            self.learning_rate,
            state = self.state
        )
        # --- optimizers ---
        self.optimizer = tf.keras.optimizers.Adam(
            learning_rate = self.learning_rate(),
            clipnorm      = self.clipnorm
        )
        # track weights
        ub.utils.set_optimizer_params(
            self.optimizer,
            self.agent.trainable_variables
        )
        
    def setup(self):
        '''(Override) Setup models, optimizers and schedulers for training'''
        self._setup_buffer()
        self._setup_model()
        self._setup_optimizer()

    # --- forward group ---
    @tf.function
    def call(self, inputs, 
                   proc_obs: bool = True,
                   proc_act: bool = False,
                   det:      bool = False,
                   training: bool = True):
        '''(Override) See Agent.call'''
        return self.agent(
            inputs,
            proc_obs = proc_obs,
            proc_act = proc_act,
            det      = det,
            training = training
        )

    def predict(self, inputs,
                      proc_obs: bool = True,
                      proc_act: bool = True,
                      det:      bool = True):
        '''(Override) See Agent.predict'''
        return self.agent.predict(
            inputs,
            proc_obs = proc_obs,
            proc_act = proc_act,
            det      = det
        )
    
    # --- collect group ---
    def _collect_step(self, obs):
        '''(Override) Collect one sample

        Args:
            obs (np.ndarray, optional): Current observations.
        
        Returns:
            np.ndarray: next observations
        '''
        # sample actions from policy
        rawact, values, log_probs = self.call(obs, proc_act=False, det=False)
        act = self.agent.proc_action(rawact)
        # convert to np.ndarray
        rawact    = np.asarray(rawact)
        act       = np.asarray(act)
        values    = np.asarray(values)    # (b,)
        log_probs = np.asarray(log_probs) # (b,)
        # step environment
        next_obs, rew, done, infos = self.env.step(act)
        # add to buffer
        self.buffer.add(
            obs  = obs,
            act  = rawact,
            done = done,
            rew  = rew,
            val  = values,
            logp = log_probs,
        )
        return next_obs
    
    def run(self, steps, obs=None):
        '''(Override) Run rollouts collection procedure'''
        # reset buffer
        self.buffer.reset()
        # collect rollouts
        obs = self.collect(steps, obs=obs)
        # compute gae
        adv = ub.data.compute_advantage(
            rew        = self.buffer.data['rew'],
            val        = self.buffer.data['val'],
            done       = self.buffer.data['done'],
            gamma      = self.gamma,
            gae_lambda = self.gae_lambda
        )
        self.buffer.data['adv'] = adv
        # ready for sampling
        self.buffer.make()
        return obs

    # --- train group ---
    def policy_loss(self, adv, logp, old_logp):
        '''Compute policy loss (clipped surrogate loss)
        Dual-clip from: https://arxiv.org/abs/1912.09729
        Mastering Complex Control in MOBA Games with Deep Reinforcement Learning

        Args:
            adv (tf.Tensor): Batch advantages (GAE), shape (b,), tf.float32
            logp (tf.Tensor): Batch action log likelihoods, shape (b,),
                tf.float32
            old_logp (tf.Tensor): Batch old action log likelihoods, shape 
                (b,), tf.float32

        Returns:
            tf.Tensor: policy loss, tf.float32
        '''
        adv = (adv - tf.math.reduce_mean(adv)) / tf.math.reduce_std(adv)
        # policy ratio
        ratio = tf.exp(logp - old_logp)
        # clipped surrogate loss
        loss1 = adv * ratio
        loss2 = adv * tf.clip_by_value(ratio, 1.-self.policy_clip,
                                              1.+self.policy_clip)
        if self.dual_clip is not None:
            clip1 = tf.minimum(loss1, loss2)
            clip2 = tf.maximum(clip1, self.dual_clip * adv)
            pi_loss = tf.where(adv < 0., clip2, clip1)
        else:
            pi_loss = tf.minimum(loss1, loss2)
        return -tf.math.reduce_mean(pi_loss)

    def value_loss(self, adv, val, old_val):
        '''Compute value loss

        Y = A(s) + V_old(s)
        loss = (V(s)-Y)^2

        Args:
            adv (tf.Tensor): Batch advantages, shape (b,)
            val (tf.Tensor): Batch value predictions, shape (b,)
            old_val (tf.Tensor): Batch old value predictions, shape (b,)

        Returns:
            tf.Tensor: value loss, shape (b,)
        '''
        y = adv + old_val
        if self.value_clip is None:
            vf_loss = tf.math.square(val-y)
        else:
            v = old_val + tf.clip_by_value(val-old_val, -self.value_clip,
                                                         self.value_clip)
            loss1 = tf.math.square(val-y)
            loss2 = tf.math.square(v-y)
            vf_loss = tf.maximum(loss1, loss2)
        return tf.math.reduce_mean(vf_loss)

    def reg_loss(self, var_list):
        '''Regularization loss

        Args:
            var_list (list): A list of variables need to be regularized.

        Returns:
            tf.Tensor: regularization loss.
        '''
        reg = tf.keras.regularizers.L2(self.reg_coef)
        return tf.math.add_n([reg(var) for var in var_list] + [0.0])

    @tf.function
    def _train_model(self, batch):
        '''Perform one gradient update'''
        vars = self.agent.trainable_variables
        with tf.GradientTape() as tape:
            tape.watch(vars)
            # forward
            dist = self.agent.call_policy(batch['obs'], training=True)
            val  = self.agent.call_value(batch['obs'], training=True)
            logp = dist.log_prob(batch['act'])
            ent  = dist.entropy()
            kl   = 0.5 * tf.math.reduce_mean(
                            tf.math.square(batch['logp'] - logp))
            # compute policy/value loss
            pi_loss  = self.policy_loss(batch['adv'], logp, batch['logp'])
            vf_loss  = self.value_loss(batch['adv'], val, batch['val'])
            ent_loss = -tf.math.reduce_mean(ent)
            reg_loss = self.reg_loss(vars)
            # aggregate losses
            loss = (pi_loss + self.vf_coef * vf_loss 
                            + self.ent_coef * ent_loss + reg_loss)
        # perform gradients
        grads = tape.gradient(loss, vars)
        self.optimizer.apply_gradients(zip(grads, vars))
        return {
            'loss':     loss,
            'pi_loss':  pi_loss,
            'vf_loss':  vf_loss,
            'ent_loss': ent_loss,
            'reg_loss': reg_loss,
            'kl':       kl
        }, kl

    def _train_step(self, batch):
        '''(Override) Train one step'''
        # update learning rate
        lr = tf.convert_to_tensor(self.learning_rate())
        self.optimizer.lr.assign(lr)
        # train model
        losses, kl = self._train_model(batch)
        return losses, kl

    def train(self, batch_size, subepochs, **kwargs):
        '''(Override) Train one epoch'''
        all_losses = []
        for subepoch in range(subepochs):
            all_kls = []
            # train one subepoch
            for batch in self.sampler(batch_size):
                # train one step
                losses, kl = self._train_step(batch)
                all_losses.append(losses)
                all_kls.append(kl)
                # update state
                self.num_gradsteps += 1
            # update states
            self.num_subepochs += 1
            # compute average kl for early stop
            if self.target_kl is not None:
                kl = np.mean(np.hstack(np.asarray(all_kls)))
                if kl > 1.5 * self.target_kl:
                    self.LOG.warning(f'Early stopping at epoch {self.num_epochs} '
                        f'due to reaching max kl: {kl:.6f}')
                    break
        # aggregate losses
        all_losses = ub.utils.flatten_dicts(all_losses)
        m_losses = {}
        for name, losses in all_losses.items():
            m_losses[name] = np.mean(np.hstack(np.asarray(losses)))
        return m_losses

    def get_config(self):
        config = super().get_config()
        config.update({
            'learning_rate': self.learning_rate,
            'gamma':         self.gamma,
            'gae_lambda':    self.gae_lambda,
            'policy_clip':   self.policy_clip,
            'value_clip':    self.value_clip,
            'dual_clip':     self.dual_clip,
            'ent_coef':      self.ent_coef,
            'vf_coef':       self.vf_coef,
            'reg_coef':      self.reg_coef,
            'clipnorm':      self.clipnorm,
            'target_kl':     self.target_kl,
            'share_net':     self.share_net,
            'force_mlp':     self.force_mlp,
            'mlp_units':     self.mlp_units
        })
        return config


# --- built in ---
import os
import sys
import enum
import time
import logging
import multiprocessing

from typing import Union

# --- 3rd party ---
import gym
import numpy as np
import cloudpickle

# --- my module ---
from unstable_baselines.lib import utils as ub_utils
from unstable_baselines.lib.envs.vec import base as vec_base

__all__ = [
    'SubprocVecEnv'
]


class CloudpickleWrapper():
    def __init__(self, **kwargs):
        self.kwargs = kwargs

    def __getattr__(self, key):
        return self.kwargs.get(key)

    def __getstate__(self):
        return cloudpickle.dumps(self.kwargs)

    def __setstate__(self, kwargs):
        self.kwargs = cloudpickle.loads(kwargs)

# Commands
class CMD:
    getattr = 1
    setattr = 2
    reset   = 3
    step    = 4
    seed    = 5
    render  = 6
    close   = 7

def _subproc_worker(_p, p, param_wrapper):
    _p.close()
    env = param_wrapper.fn()
    auto_reset = param_wrapper.auto_reset
    try:
        while True:
            try:
                cmd, data = p.recv()
            except EOFError: # the pipe has been closed
                p.close()
                break
            if cmd == CMD.getattr:
                p.send(getattr(env, data[0], None))
            elif cmd == CMD.setattr:
                p.send(setattr(env, data[0], data[1]))
            elif cmd == CMD.reset:
                p.send(env.reset(**data[0]))
            elif cmd == CMD.step:
                obs, rew, done, info = env.step(data[0])
                if auto_reset and done:
                    obs = env.reset()
                p.send((obs, rew, done, info))
            elif cmd == CMD.seed:
                p.send(env.seed(data[0]))
            elif cmd == CMD.render:
                p.send(env.render(**data[0]))
            elif cmd == CMD.close:
                p.send(env.close())
                p.close()
                break
            else:
                p.close()
                raise NotImplementedError
    except KeyboardInterrupt:
        p.close()

class SubprocEnvWorker(vec_base.BaseEnvWorker):
    def __init__(self, env_fn, auto_reset: bool):
        methods = multiprocessing.get_all_start_methods()
        start_method = 'spawn'
        if 'forkserver' in methods:
            start_method = 'forkserver'
        ctx = multiprocessing.get_context(start_method)
        self.p, _p = ctx.Pipe()
        args = (
            self.p, _p, CloudpickleWrapper(fn=env_fn, auto_reset=auto_reset)
        )
        self.process = ctx.Process(target=_subproc_worker, args=args, daemon=True)
        self.process.start()
        self._waiting_cmd = None
        _p.close()
        super().__init__(env_fn, auto_reset)
        
    def getattr(self, attrname: str):
        return self._cmd(CMD.getattr, attrname)

    def setattr(self, attrname: str, value):
        return self._cmd(CMD.setattr, attrname, value)

    def reset(self, **kwargs):
        return self._cmd(CMD.reset, kwargs)

    def step_async(self, act):
        return self._cmd(CMD.step, act, wait=False)

    def step_wait(self):
        return self._wait(CMD.step)

    def seed(self, seed):
        super().seed(seed)
        return self._cmd(CMD.seed, seed)

    def render(self, **kwargs):
        return self._cmd(CMD.render, kwargs)

    def close_async(self):
        return self._cmd(CMD.close, wait=False)

    def close_wait(self):
        return self._wait(CMD.close, timeout=1)
    
    def _cmd(self, cmd, *args, wait=True):
        #TODO: find a more reliable way
        if self._waiting_cmd and cmd != CMD.close:
            raise RuntimeError(f'Another command {cmd} was sent when '
                f'waiting for the reply {self._waiting_cmd}')
        self.p.send([cmd, args])
        self._waiting_cmd = cmd # marked as waiting reply
        if wait:
            return self._wait(cmd)

    def _wait(self, cmd, timeout=None):
        if self._waiting_cmd != cmd:
            raise RuntimeError
        res = None
        if self.p.poll(timeout):
            res = self.p.recv()
        self._waiting_cmd = None #unmarked
        return res


class SubprocVecEnv(vec_base.BaseVecEnv):
    def __init__(self,
        env_fns: list,
        rms_norm: Union[str, bool, ub_utils.RMSNormalizer] = False,
        auto_reset: bool = True,
    ):
        super().__init__(env_fns, SubprocEnvWorker, rms_norm, auto_reset)


# --- built in ---
import os
import re
import abc
import sys
import json
import time
import base64
import random
import datetime
import itertools
import contextlib

from collections import OrderedDict

# --- 3rd party ---
import gym 
import cloudpickle

import numpy as np
import tensorflow as tf

from tensorflow.python.lib.io import file_io

# --- my module ---

__all__ = [
    'spaces',
    'NormalActionNoise',
    'StateObject',
    'RunningMeanStd',
    'set_seed',
    'normalize',
    'denormalize',
    'stack_obs',
    'soft_update',
    'is_image_space',
    'flatten_dicts',
    'is_json_serializable',
    'to_json_serializable',
    'from_json_serializable',
    'input_tensor',
    'set_optimizer_params',
    'get_tensor_ndims',
    'flatten_dims',
    'run_eagerly',
    'safe_json_dumps',
    'safe_json_loads',
    'safe_json_dump',
    'safe_json_load',
    'iter_nested',
    'map_nested',
    'iter_nested_space',
    'map_nested_space',
    'iter_nested_tuple',
    'map_nested_tuple',
    'nested_to_numpy',
    'unpack_structure',
    'pack_sequence',
    'flatten_space'
]



# === Const params ===

SERIALIZATION_KEY='#SERIALIZED'

# Extra space profiles
class spaces:
    All = [gym.spaces.Box, gym.spaces.Discrete,
           gym.spaces.MultiBinary, gym.spaces.MultiDiscrete,
           gym.spaces.Tuple, gym.spaces.Dict]
    NonNested = [gym.spaces.Box, gym.spaces.Discrete,
                 gym.spaces.MultiBinary, gym.spaces.MultiDiscrete]
    Nested = [gym.spaces.Tuple, gym.spaces.Dict]

# === Action noises ===
#TODO: Move action noises to environment wrapper
class NormalActionNoise():
    def __init__(self, mean, scale):
        self.mean = mean
        self.scale = scale

    def __call__(self, shape=None):
        return np.random.normal(self.mean, self.scale, size=shape)

    def __repr__(self):
        return f'NormalActionNoise(mean={self.mean}, scale={self.scale})'

    def reset(self):
        pass

# === StateObject ===
class StateObject(dict):
    '''An object-like dictionary that you can get/set
    items/attributes by either __getitem__/__setitem__
    or __getattr__/__setattr__.
    For example:
    >>> d = StateObject()
    >>> d['abc'] = 10
    >>> print(d['abc'])
    This is equivalent to
    >>> d.abc = 10
    >>> print(d.abc)
    Also, StateObject can be serialized by JSON by calling
    `tostring` and `fromstring`.
    '''
    def __new__(cls, *args, **kwargs):
        self = super().__new__(cls, *args, **kwargs)
        self.__dict__ = self
        return self
    
    def tostring(self, indent=None):
        return safe_json_dumps(self, indent=indent)

    @classmethod
    def fromstring(cls, string):
        self = StateObject()
        self.update(safe_json_loads(string))
        return self

# === RunningMeanStd ===
class RunningMeanStd(StateObject):
    def __init__(self, mean: float = 0.0, 
                       std:  float = 1.0):
        self.mean  = mean
        self.var   = std ** 2.0
        self.count = 0

    def update(self, x: np.ndarray):
        batch_mean = np.mean(x, axis=0)
        batch_var = np.var(x, axis=0)
        batch_count = len(x)

        delta = batch_mean - self.mean
        total_count = self.count + batch_count

        new_mean = self.mean + delta * batch_count / total_count
        m_a = self.var * self.count
        m_b = batch_var * batch_count
        m_2 = m_a + m_b + delta ** 2 * self.count * batch_count / total_count
        new_var = m_2 / total_count

        self.mean, self.var = new_mean, new_var
        self.count = total_count

# === RMS normalizer ===
class RMSNormalizer():
    def __init__(self,
        space:    gym.Space = None,
        enable:        bool = None,
        fixed:         bool = None,
        rms: RunningMeanStd = None,
    ):
        '''A RunningMeanStd normalizer for non-nested spaces.
        This class provides an interface for `VecEnv`s to perform
        running mean std observation normalization. Note that this
        normalizer only supports non-nested spaces. 
        (e.g. gym.spaces.Box)

        Args:
            space (gym.Space): Gym space, expecting a gym.spaces.Boxl, float32.
            enable (bool, optional): Enable normalization. Default to True if
                `space` can be normalized. (Box, float32, not image)
            fixed (bool, optional): Whether to fix RMS value.
                Defaults to False.
            rms (RunningMeanStd, optional): RMS init value. Defaults to None.
        '''
        self.space    = space
        self._enable  = enable
        self._fixed   = fixed
        self._rms     = rms
        self.is_setup = False
        if space is not None:
            self.setup(space)

    def setup(self, space: gym.Space=None):
        '''Setup normalizer'''
        assert not self.is_setup, 'This RMSNormalizer is already been setup'
        self.space    = space
        self._enable  = self._get_enable_opt(self._enable)
        self._fixed   = self._get_fixed_opt(self._fixed)
        self._rms     = self._get_rms_opt(self._rms)
        self.is_setup = True
        return self

    @property
    def rms(self):
        return self._rms

    @property
    def enabled(self):
        return self._enable

    @enabled.setter
    def enabled(self, value: bool):
        self._enable = bool(value)

    @property
    def fixed(self):
        return self._fixed

    @fixed.setter
    def fixed(self, value: bool):
        self._fixed = bool(value)

    def normalize(self, x: np.ndarray, clip_max: float=10.0):
        '''Normalize x with mean and var'''
        # this magic number is from openai baselines
        # see baselines/common/vec_env/vec_normalize.py#L10
        if not self.enabled:
            return x
        rms = self.rms
        eps = np.finfo(np.float32).eps.item()
        x = (x-rms.mean)/np.sqrt(rms.var+eps)
        x = np.clip(x, -clip_max, clip_max)
        return x

    def __call__(self, x:np.ndarray, clip_max: float=10.0):
        '''Shortcut for `normalize()`'''
        return self.normalize(x, clip_max)

    def update(self, x:np.ndarray):
        '''Update RMS value'''
        if self.fixed:
            return
        if len(x.shape) == len(self.space.shape):
            # one sample, expand batch dim
            x = np.expand_dims(x, axis=0)
        self.rms.update(x)

    def load(self, path: str):
        '''Load RMS value from `path`
        Usage:
            >>> rms = RMSNormalizer(
            ...     env.observation_space,
            ...     enable = True,
            ...     fixed = True
            ... ).load('/saved_rms.json')
        '''
        d = safe_json_load(path)
        assert set(['mean','var','count']) == set(d.keys())
        #NOTE: RunningMeanStd is a dict type
        dict.update(self.rms, d)
        return self

    def save(self, path: str):
        '''Save RMS value to `path`'''
        safe_json_dump(path, self.rms)

    def _get_enable_opt(self, opt):
        if opt is None:
            if self.space is None:
                # return True for unknown space
                return True
            # Enabled if space is a Box but not image.
            return (not is_image_space(self.space)
                and isinstance(self.space, gym.spaces.Box)
                and np.dtype(self.space.dtype) == np.float32)
        return bool(opt)
    
    def _get_fixed_opt(self, opt):
        return (not self._enable) or (opt is True)

    def _get_rms_opt(self, opt):
        if isinstance(opt, RunningMeanStd):
            return opt
        return RunningMeanStd()


class NestedRMSNormalizer(RMSNormalizer):
    '''A Universal RunningMeanStd normalizer for both nested and 
    non-nested spaces.
    Since this normalizer may cause performance issue, we don't
    encourage you to use this normalizer. Instead, you can implement
    a custom RMSNormalizer for your custom space.
    '''
    def __init__(self, 
        space:    gym.Space,
        enable:        bool = None,
        fixed:         bool = None,
        rms: RunningMeanStd = None,
    ):
        print('WARN: We don\'t encourage you to use this RMS '
                'normalizer for nested spaces (Dict/Tuple). '
                'See ub.lib.utils.NestedRMSNormalizer')
        #TODO
        raise NotImplementedError
    #     assert isinstance(space, gym.Space)
    #     self.space = space

    #     pyspace = nested_iter_space(self.space, lambda sp:sp)
    #     self._pyspace = pyspace
    #     self._enable_norm = self._get_enable_norm_opt(enable_norm)
    #     self._update_norm = self._get_update_norm_opt(update_norm)
    #     self._norm_rms    = self._get_norm_rms_opt(norm_rms)

    # def _get_enable_norm_opt(self, opt):
    #     nest_opt = self._form_nested_opt(opt, arg_name='enable_norm')
    #     return nested_iter_tuple((nest_opt, self._pyspace),
    #                 lambda t: (t[0] is not False) and is_image_obs(t[1]))

    # def _get_update_norm_opt(self, opt):
    #     nest_opt = self._form_nested_opt(opt, arg_name='update_norm')
    #     return nested_iter_tuple((nest_opt, self._enable_norm),
    #             lambda t: (t[0] is not False) and t[1])

    # def _get_norm_rms_opt(self, opt):


    # def _form_nested_opt(self, opt, arg_name):
    #     if isinstance(opt, (list, tuple, dict)):
    #         try:
    #             nested_iter_tuple((self._pyspace, opt), 
    #                     lambda _: 0)
    #         except:
    #             raise ValueError(f'If `{arg_name}` is a nested data, '
    #                 'it must have a same structure as `space`')
    #         nest_opt = opt
    #     else:
    #         nest_opt = nested_iter(self._pyspace, lambda _: opt)
    #     return nest_opt


# === Utils ===

def set_seed(seed):
    tf.random.set_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


def normalize(x, low, high, nlow=0.0, nhigh=1.0):
    '''
    Normalize x from [low, high] to [nlow, nhigh]
    '''
    return ((nhigh-nlow)/(high-low)) * (x-low) + nlow

def denormalize(x, low, high, nlow=0.0, nhigh=1.0):
    '''
    Denormalize x from [nlow, nhigh] to [low, high]
    '''
    return ((high-low)/(nhigh-nlow)) * (x-nlow) + low


def stack_obs(obs, space):
    if not isinstance(obs, (list, tuple)):
        raise ValueError('Expecting a list or tuple of observations')
    if len(obs) <= 0:
        raise ValueError('Need observations from at least one environment')
    if isinstance(space, gym.spaces.Dict):
        if not isinstance(space.spaces, OrderedDict):
            raise ValueError('Dict space must have ordered subspaces')
        if not isinstance(obs[0], dict):
            raise ValueError('non-dict observation for environment with '
                    'Dict observation space')
        return OrderedDict([
            (k, stack_obs([o[k] for o in obs], space.spaces[k]))
            for k in space.spaces.keys()
        ])
    elif isinstance(space, gym.spaces.Tuple):
        if not isinstance(obs[0], tuple):
            raise ValueError('non-tuple observation for environment '
                    'with Tuple observation space')
        return tuple([
            stack_obs([o[i] for o in obs], space.spaces[i])
            for i in range(len(space.spaces))
        ])
    else:
        return np.stack(obs)

def is_image_space(space):
    return (isinstance(space, gym.spaces.Box)
                and np.dtype(space.dtype) == np.uint8
                and len(space.shape) == 3)

def flatten_dicts(dicts: list):
    '''Flatten a list of dicts

    Args:
        dicts (list): list of dicts

    Returns:
        dict: flattened dict
    '''
    agg_dict = {}
    for d in dicts:
        for k, v in d.items():
            agg_dict.setdefault(k, []).append(v)
    return agg_dict

def is_bounded(space):
    if isinstance(space, gym.spaces.Box):
        return (not np.any(np.isinf(space.low))
                and not np.any(np.isinf(space.high))
                and np.all(space.high-space.low > 0.))
    return True

# === TensorFlow utils ===

def input_tensor(space, batch_size=1):
    _tensor_op = lambda sp: (tf.zeros((batch_size, *sp.shape), 
                                       dtype=sp.dtype))
    return map_nested_space(space, _tensor_op)

def set_optimizer_params(optimizer, param_list):
    # Set optimizer tracking parameters
    if len(param_list) > 0:
        with tf.name_scope(optimizer._name):
            with tf.init_scope():
                optimizer._create_all_weights(param_list)
    return optimizer

def preprocess_observation(inputs, obs_space, dtype=tf.float32):
    '''Preprocess non-float observations
    If the input's dtype is not float32 or float64 we normalize it.
    '''
    inputs = tf.convert_to_tensor(inputs)
    # Do nothing if input is a float
    if tf.as_dtype(inputs.dtype) in [tf.float32, tf.float64]:
        return tf.cast(inputs, dtype=dtype)
    # Normalize
    if isinstance(obs_space, gym.spaces.Box):
        inputs = tf.cast(inputs, dtype=dtype)
        if is_bounded(obs_space):
            low    = tf.constant(obs_space.low, dtype=dtype)
            high   = tf.constant(obs_space.high, dtype=dtype)
            inputs = normalize(inputs, low, high, 0., 1.)
    elif isinstance(obs_space, gym.spaces.Discrete):
        depth  = tf.constant(obs_space.n, dtype=tf.int32)
        inputs = tf.one_hot(inputs, depth=depth)
    elif isinstance(obs_space, gym.spaces.MultiDiscrete):
        # inputs = [3, 5] obs_space.nvec = [4, 7]
        # [0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0]
        nvec   = tf.constant(obs_space.nvec, dtype=tf.int32)
        inputs = tf.concat([
                tf.one_hot(inputs[..., idx], depth=nvec[idx])
                for idx in range(inputs.shape[-1])
            ], axis=-1)
    elif isinstance(obs_space, gym.spaces.MultiBinary):
        pass
    else:
        raise NotImplementedError("Preprocessing not implemented for "
                                f"{type(obs_space)}")

    return tf.cast(inputs, dtype=dtype)

def get_tensor_ndims(tensor: tf.Tensor):
    tensor = tf.convert_to_tensor(tensor)
    ndims = tensor.get_shape().ndims or tf.rank(tensor)
    return ndims


def flatten_dims(tensor: tf.Tensor, begin, end=None):
    '''Collapse dims

    Flatten tensor from dim `begin` to dim `end`

    For example:
        >>> tensor = tf.zeros((2, 3, 4))
        >>> print(collapse_dims(tensor, 1).shape)
        (2, 12)

    Args:
        tensor (tf.Tensor): [description]
        begin ([type]): [description]
        end ([type], optional): [description]. Defaults to None.

    Returns:
        [type]: [description]
    '''    
    tensor = tf.convert_to_tensor(tensor)
    if end is None:
        flat_shape = tf.concat((
            tf.shape(tensor)[:begin],
            [tf.math.reduce_prod(tf.shape(tensor)[begin:])],
        ), axis=0)
    else:
        flat_shape = tf.concat((
            tf.shape(tensor)[:begin],
            [tf.math.reduce_prod(tf.shape(tensor)[begin:end])],
            tf.shape(tensor)[end:]
        ), axis=0)

    return tf.reshape(tensor, flat_shape)

def soft_update(target_vars, source_vars, polyak=0.005):
    '''Perform soft updates

    target_var = (1-polyak) * target_var + polyak * source_var

    Args:
        target_vars (list): A list of tf.Variable update to.
        source_vars (list): A list of tf.Variable update from. 
            The length must be equal to target_vars.
        polyak (float, optional): Smooth rate. Defaults to 0.005.
    '''
    if len(target_vars) != len(source_vars):
        raise ValueError('Length does not match, '
            f'got {len(target_vars)} and {len(source_vars)}')

    for (tar_var, src_var) in zip(target_vars, source_vars):
        tar_var.assign((1.-polyak) * tar_var + polyak * src_var)

def broadcast_shape(x_shape, y_shape):
    x_shape_static = tf.get_static_value(x_shape)
    y_shape_static = tf.get_static_value(y_shape)
    if (x_shape_static is None) or (y_shape_static is None):
        return tf.broadcast_dynamic_shape(x_shape, y_shape)
    return tf.broadcast_static_shape(
        tf.TensorShape(x_shape_static), tf.TensorShape(y_shape_static))

@contextlib.contextmanager
def run_eagerly(enabled=True):
    orig_conf = tf.executing_eagerly()
    tf.config.run_functions_eagerly(enabled)
    try:
        yield
    finally:
        tf.config.run_functions_eagerly(orig_conf)

# === JSON utils ===

def is_json_serializable(obj):
    '''
    Check if the object is json serializable
    '''
    try:
        json.dumps(obj, ensure_ascii=False)
        return True
    except:
        return False

def to_json_serializable(obj):
    '''Convert any object into json serializable object

    Args:
        obj (Any): Any object.
    '''
    if not is_json_serializable(obj):
        encoded_obj = base64.b64encode(
            cloudpickle.dumps(obj)).decode()
        encoded_obj = {SERIALIZATION_KEY: encoded_obj}
    else:
        encoded_obj = obj

    return encoded_obj

def from_json_serializable(encoded_obj):
    '''Convert any serializable object into original object

    Args:
        encoded_obj (Any): Serializable object
    '''
    if (isinstance(encoded_obj, dict) 
            and SERIALIZATION_KEY in encoded_obj.keys()):
        obj = encoded_obj[SERIALIZATION_KEY]
        obj = cloudpickle.loads(
            base64.b64decode(obj.encode()))
    else:
        obj = encoded_obj
    return obj

def safe_json_dumps(obj, 
                    indent=2, 
                    ensure_ascii=False, 
                    default=to_json_serializable, 
                    **kwargs):
    string = json.dumps(obj, indent=indent, 
                        ensure_ascii=ensure_ascii, 
                        default=default,
                        **kwargs)
    return string

def safe_json_loads(string, 
                    object_hook=from_json_serializable, 
                    **kwargs):
    obj = json.loads(string, 
                    object_hook=object_hook, 
                    **kwargs)
    return obj


def safe_json_dump(filepath,
                    obj,
                    **kwargs):
    string = safe_json_dumps(obj, **kwargs)
    file_io.atomic_write_string_to_file(filepath, string)


def safe_json_load(filepath,
                    **kwargs):
    obj = None
    if file_io.file_exists(filepath):
        string = file_io.read_file_to_string(filepath)
        obj = safe_json_loads(string, **kwargs)

    return obj

# === nested ops ===

def iter_nested(data, sortkey=False):
    '''Iterate over nested data structure
    Note: Use `tuple` instead of `list`. A list type
    object is treated as an item.

    For example:
    >>> data = {'a': (1, 2), 'b': 3}
    >>> list(v for v in iter_nested(data))
    [1, 2, 3]

    Args:
        data (tuple or dict): A nested data.
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    def _inner_iter_nested(data):
        if isinstance(data, dict):
            keys = sorted(data.keys()) if sortkey else data.keys()
            for k in keys:
                yield from _inner_iter_nested(data[k])
        elif isinstance(data, tuple):
            for v in data:
                yield from _inner_iter_nested(v)
        else:
            yield data
    return _inner_iter_nested(data)

def map_nested(data, op, *args, sortkey=False, **kwargs):
    '''A nested version of map function
    NOTE: Use `tuple` instead of `list`. A list type 
    object is treated as an item.

    Args:
        data (tuple or dict): A nested data
        op (function): A function operate on each data
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    if not callable(op):
        raise ValueError('`op` must be a callable')

    def _inner_map_nested(data):
        if isinstance(data, dict):
            keys = sorted(data.keys()) if sortkey else data.keys()
            return {k: _inner_map_nested(data[k])
                        for k in keys}
        elif isinstance(data, tuple):
            return tuple(_inner_map_nested(v)
                            for v in data)
        else:
            return op(data, *args, **kwargs)
    return _inner_map_nested(data)

def iter_nested_space(space, sortkey=False):
    '''Iterate over nested gym space. Similar to iter_nested
    but it's for gym spaces.

    Args:
        space (gym.Space): Nested or non-nested gym space
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    def _inner_iter_nested(space):
        if isinstance(space, gym.spaces.Dict):
            spaces = space.spaces
            keys = sorted(spaces.keys()) if sortkey else spaces.keys()
            for k in keys:
                yield from _inner_iter_nested(spaces[k])
        elif isinstance(space, gym.spaces.Tuple):
            for v in space.spaces:
                yield from _inner_iter_nested(v)
        else:
            yield space
    return _inner_iter_nested(space)

def map_nested_space(space, op, *args, sortkey=False, **kwargs):
    '''A nested version of map function. Similar to map_nested
    but it's for gym spaces.

    Args:
        space (gym.Space): Nested or non-nested gym space
        op (callable): A function operate on each data
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    if not callable(op):
        raise ValueError('`op` must be a callable')

    def _inner_map_nested(space):
        if isinstance(space, gym.spaces.Dict):
            spaces = space.spaces
            keys = sorted(spaces.keys()) if sortkey else spaces.keys()
            return {k: _inner_map_nested(spaces[k])
                        for k in keys}
        elif isinstance(space, gym.spaces.Tuple):
            return tuple(_inner_map_nested(v)
                            for v in space.spaces)
        else:
            return op(space, *args, **kwargs)
    return _inner_map_nested(space)

def iter_nested_tuple(data_tuple, sortkey=False):
    '''Iterate over a tuple of nested structures. Similar to iter_nested
    but it iterates each of each nested data in the input tuple.
    For example:
    >>> a = {'x': 1, 'y': (2, 3)}
    >>> b = {'u': 4, 'v': (5, 6)}
    >>> list(iter_nested_tuple((a, b)))
    [(1, 4), (2, 5), (3, 6)]

    Args:
        data_tuple (tuple): A tuple of nested data.
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    if not isinstance(data_tuple, tuple):
        raise TypeError('`data_tuple` only accepts tuple, '
                f'got {type(data_tuple)}')
    def _inner_iter_nested(data_tuple):
        if isinstance(data_tuple[0], dict):
            keys = data_tuple[0].keys()
            keys = sorted(keys) if sortkey else keys
            for k in keys:
                yield from _inner_iter_nested(
                        tuple(data[k] for data in data_tuple))
        elif isinstance(data_tuple[0], tuple):
            for k in range(len(data_tuple[0])):
                yield from _inner_iter_nested(
                    tuple(data[k] for data in data_tuple))
        else:
            yield data_tuple
    return _inner_iter_nested(data_tuple)

def map_nested_tuple(data_tuple, op, *args, sortkey=False, **kwargs):
    '''A nested version of map function. Similar to map_nested
    but it iterates each of each nested data in the input tuple.

    Args:
        data_tuple (tuple): A tuple of nested data.
        op (callable): A function operate on each data
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    if not callable(op):
        raise ValueError('`op` must be a callable')
    if not isinstance(data_tuple, tuple):
        raise TypeError('`data_tuple` only accepts tuple, '
                f'got {type(data_tuple)}')
    def _inner_map_nested(data_tuple):
        if isinstance(data_tuple[0], dict):
            keys = data_tuple[0].keys()
            keys = sorted(keys) if sortkey else keys
            return {k: _inner_map_nested(
                        tuple(data[k] for data in data_tuple))
                    for k in keys}
        elif isinstance(data_tuple[0], tuple):
            return tuple(_inner_map_nested(
                        tuple(data[idx] for data in data_tuple))
                    for idx in range(len(data_tuple[0])))
        else:
            return op(data_tuple, *args, **kwargs)
    return _inner_map_nested(data_tuple)

def nested_to_numpy(data, sortkey=False):
    '''Convert all items in a nested data into 
    numpy arrays

    Args:
        data (dict, tuple): A nested data
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.

    Returns:
        dict, tuple: A nested data same as `data`
    '''    
    op = lambda arr: np.asarray(arr)
    return map_nested(data, op, sortkey=sortkey)

def unpack_structure(data, sortkey=False):
    '''Extract structure and flattened data from a nested data
    For example:
        >>> data = {'a': 'abc', 'b': (2.0, [3, 4, 5])}
        >>> struct, flat_data = extract_struct(data)
        >>> flat_data
        ['abc', 2.0, [3, 4, 5]]
        >>> struct
        {'a': 0, 'b': (1, 2)}
    
    Args:
        data (dict, tuple): A nested data
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.

    '''
    _count_op = lambda v, c: next(c)
    counter = itertools.count(0)
    struct = map_nested(data, _count_op, counter, sortkey=sortkey)
    size = next(counter)
    flat_data = [None] * size
    def _flat_op(ind_and_data, flat_data):
        ind, data = ind_and_data
        flat_data[ind] = data
    map_nested_tuple((struct, data), _flat_op, flat_data, sortkey=sortkey)
    return struct, flat_data

def pack_sequence(struct, flat_data, sortkey=False):
    '''An inverse operation of `extract_structure`

    Args:
        struct: A nested structure each data field contains
            an index of elements in `flat_data`
        flat_data (list): flattened data.
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.
    '''
    _struct_op = lambda ind, flat: flat[ind]
    data = map_nested(struct, _struct_op, flat_data, sortkey=sortkey)
    return data

def flatten_space(space: gym.Space, sortkey=False):
    '''Flatten gym's nested space, like gym.Dict or gym.Tuple,
    to a tuple of spaces.

    Args:
        space (gym.Space): Nested or non-nested space.
        sortkey (bool): Whether to sort dict's key. Defaults
            to False.

    Returns:
        tuple: Flattened gym spaces.
    '''
    return tuple(v for v in iter_nested_space(space, sortkey=sortkey))


# --- built in ---
import os
import abc
import sys
import math
import time
import logging

# --- 3rd party ---
import numpy as np

# --- my module ---
from unstable_baselines.lib import utils as ub_utils

__all__ = [
    'SegmentTree',
    'ReplayBuffer',
    'DynamicBuffer',
    'UniformSampler',
    'PrioritizedSampler',
    'PriorSampler', # alias: PrioritizedSampler
    'PermuteSampler',
    'compute_nstep_rew',
    'compute_advantage'
]

# === Segment Tree ===

class SegmentTree(metaclass=abc.ABCMeta):
    def __init__(self, size: int):
        '''An implementation of segment tree used to efficiently O(logN)
        compute the sum of a query range [start, end)

        Args:
            size (int): Number of elements.
        '''
        assert isinstance(size, int) and size > 0
        base = 1<<(size-1).bit_length()
        self._size = size
        self._base = base
        self._value = np.zeros([base * 2], dtype=np.float64)
    
    def __getitem__(self, key: np.ndarray):
        # formalize indices
        if isinstance(key, (int, slice)):
            key = np.asarray(range(self._size)[key], dtype=np.int64)
        else:
            key = np.asarray(key, dtype=np.int64)
        key = key % self._size + self._base
        return self._value[key]

    def __setitem__(self, key: np.ndarray, value: np.ndarray):
        self.update(key, value)

    def update(self, key: np.ndarray, value: np.ndarray):
        '''Update elements' values'''
        # formalize indices
        if isinstance(key, (int, slice)):
            key = np.asarray(range(self._size)[key], dtype=np.int64)
        else:
            key = np.asarray(key, dtype=np.int64)
        key = key % self._size + self._base
        key = key.flatten()
        value = np.asarray(value, dtype=np.float64).flatten()
        # set values
        self._value[key] = value
        # update tree (all keys have the same depth)
        while key[0] > 1:
            self._value[key>>1] = self._value[key] + self._value[key^1]
            key >>= 1

    def sum(self, start: int=None, end: int=None):
        '''Compute the sum of the given range [start, end)'''
        if (start == None) and (end == None):
            # shortcut
            return self._value[1]
        start, end, _ = slice(start, end).indices(self._size)
        start += self._base
        end += self._base
        res = 0.0
        while start < end:
            if start & 1:
                res += self._value[start]
            if end & 1:
                res += self._value[end-1]
            start = (start+1) >> 1
            end = end >> 1
        return res

    def index(self, value: np.ndarray):
        '''Return the largest index such that
        value[0:index+1].sum() < value
        '''
        assert np.min(value) >= 0.0
        assert np.max(value) < self._value[1]
        # if input is a scalar, return should be a scalar too.
        one_value = np.isscalar(value)
        # convert to 1D array
        value = np.asarray(value, dtype=np.float64)
        orig_shape = value.shape
        value = value.flatten()
        inds = np.ones_like(value, dtype=np.int64)
        # find inds (all inds have the same depth)
        while inds[0] < self._base:
            inds <<= 1
            lsum = self._value[inds]
            d = lsum < value
            value -= lsum * d
            inds += d
        inds -= self._base
        inds = inds.reshape(orig_shape)
        return inds.item() if one_value else inds

# === Replay buffer ===

class RelativeIndex():
    def __init__(self, buffer:  'BaseBuffer',
                       offsets:   tuple = None,
                       max_sizes: tuple = None):
        '''This class is used to relative indexing

        For example, if the buffer has 10 slots and the last element
        (pos) is at 6th slot, you can get the last element by absolute 
        indexing
        >>> buffer[6]
        by relative indexing (relative to `buffer.head`)
        >>> buffer.rel[-1]
        `-1` is the last element in the buffer.
        This is useful when you want to calculate N-step returns

        This also support vectorized relative indexing, for example:
        >>> rel = RelativeIndex(buffer, [1, 3, 5])
        >>> rel[-1]
        this is equivelent to
        >>> buffer[[0, 2, 4]]

        Note that relative indexing only supports int, np.ndarray and 
        slice. It does not support Ellipsis.

        Args:
            buffer (BaseBuffer): Buffer
            offsets (int, tuple, np.ndarray): The indices relative to.
                If None, it is set to `buffer.head`. Defaults to None.
            max_sizes
        '''
        assert isinstance(buffer, BaseBuffer)
        self.buffer = buffer
        if offsets is None:
            offsets = buffer.head
        if max_sizes is None:
            max_sizes = buffer.len_slots()
        self.offsets = np.index_exp[offsets]
        self.max_sizes = np.index_exp[max_sizes]
        assert len(self.offsets) == len(self.max_sizes)

    def toabs(self, key, offset, max_size):
        '''Convert to absolute indexing'''
        if key is Ellipsis:
            raise NotImplementedError("Relative indexing does not"
                "support Ellipsis")
        # formalize indices (to np.ndarray)
        if isinstance(key, (int, slice)):
            key = np.asarray(range(max_size)[key], dtype=np.int64)
        else:
            key = np.asarray(key, dtype=np.int64)
        if not np.isscalar(offset):
            # vectorized indexing (*key.shape, *offset.shape)
            key = key.reshape(key.shape + (1,)*offset.ndim)
        key = (key + offset) % max_size
        return key
    
    def cvtkeys(self, keys):
        '''Convert relative keys to absolute keys'''
        keys = np.index_exp[keys]
        min_len = min(len(self.offsets), len(keys))
        abs_keys = []
        for dim in range(min_len):
            abs_key = self.toabs(
                keys[dim],
                self.offsets[dim],
                self.max_sizes[dim]
            )
            abs_keys.append(abs_key)
        if len(keys) > min_len:
            abs_keys.extend(keys[min_len:])
        elif len(self.offsets) > min_len:
            abs_keys.extend(self.offsets[min_len:])
        return tuple(abs_keys)

    def __getitem__(self, keys):
        return self.buffer[self.cvtkeys(keys)]

    def __setitem__(self, keys, vals):
        self.buffer[self.cvtkeys(keys)] = vals

class BaseBuffer(metaclass=abc.ABCMeta):
    '''The base class of replay buffers.
    BaseBuffer stores sample in a 2D manner (slots, batch), where `slots`
    can be seen as the timesteps of each samples, and `batch` is the 
    number of samples each time user `add()` into the buffer.
    '''
    def __init__(self, size: int, batch: int=None):
        '''Create a buffer
        
        Args:
            size (int): Buffer size, capacity
            batch (int, optional): batch size of replay samples, commonly 
                referred to as number of envs in a vectorized env. This 
                value is automatically set when user first adds a batch 
                of data. Defaults to None.
        '''
        if not isinstance(size, int) or size <= 0:
            raise ValueError('`size` must be greater than 0, '
                f'got {size}')
        self._min_size = size
        self._batch = batch
        self._slots = None
        self._size  = None
        self._pos   = 0
        self._full  = False
        self._data  = None
        self.reset()
        # calculate buffer spaces
        if batch is not None:
            self._calc_space(batch)

    @property
    def capacity(self):
        '''Return the capacity of the buffer'''
        return self._size
    
    @property
    def slots(self):
        '''Return the number of total slots'''
        return self._slots
    
    @property
    def batch(self):
        '''Return batch size'''
        return self._batch
    
    @property
    def head(self):
        '''Return the index of the first slot'''
        return self._pos if self.isfull else 0

    @property
    def tail(self):
        '''Return the index of the last slot'''
        return self._pos
    
    @property
    def isnull(self):
        '''Return True, if the buffer space is not created'''
        return self._data is None

    @property
    def isfull(self):
        '''Return True, if the buffer is full'''
        return self._full

    @property
    def ready_for_sample(self):
        '''Return True, if it is ready for sampling'''
        return True

    @property
    def data(self):
        '''Retrieve data'''
        return self._data

    @property
    def rel(self):
        '''Relative indexing'''
        return RelativeIndex(self)

    def reset(self):
        '''Reset buffer'''
        self._pos    = 0
        self._full   = False
        self._data   = None

    def ravel_index(self, indices: tuple):
        '''Ravel 2D indices to 1D indices'''
        assert not self.isnull
        return np.ravel_multi_index(indices, (self.slots, self.batch))

    def unravel_index(self, indices: np.ndarray):
        '''Unravel 1D indices to 2D indices'''
        assert not self.isnull
        return np.unravel_index(indices, (self.slots, self.batch))

    def add(self, data):
        '''Add new batch data into the buffer

        Args:
            data (Any): Any nested type of data. Note that the python list
                is treated as a single element. Each data must have shape 
                (b, *)

        Return:
            tuple: 2D indices of the samples
        '''
        # count the batch dimension
        arr = next(iter(ub_utils.iter_nested(data)))
        assert not np.isscalar(arr), f'rank must be > 0'
        n_samples = len(arr)
        # prepare indices and copy data into the buffer
        cur_pos = self._pos
        self._set_data(data, indices=cur_pos)
        # update cursor position (circular buffer)
        self._pos += 1
        if self._pos >= self._slots:
            self._full = True
            self._pos = self._pos % self._slots
        return (cur_pos, np.arange(n_samples))

    def update(self, data, indices):
        '''Update buffer data'''
        self._set_data(data, indices=indices)

    def len_slots(self):
        '''Return the number of filled slots'''
        return (self._slots if self.isfull else self._pos)

    def __len__(self):
        '''Return the number of total samples'''
        if self.isnull:
            return 0
        return (self._size if self.isfull else self._pos*self._batch)

    def __getitem__(self, key):
        '''Get a slice of data
        
        Args:
            key(int, np.ndarray): Indices.

        Returns:
            Any: A slice of nested data
        '''
        return self._get_data(key)

    def __setitem__(self, key, val):
        '''Set slice of data
        
        Args:
            key (int, slice): Indices slices
            val (Any): New data
        '''
        self._set_data(val, indices=key)

    # --- private methods ---
    def _calc_space(self, batch: int):
        '''Calculate buffer spaces'''
        assert isinstance(batch, int)
        assert batch > 0
        self._batch = batch
        self._slots = math.ceil(self._min_size/batch)
        self._size  = self._slots * self._batch # true size

    def _melloc(self, data):
        '''Create spaces from the given data example'''
        def _create_space(v):
            v = np.asarray(v)
            shape = (self._slots, self._batch)
            if len(v.shape) > 1:
                shape += v.shape[1:]
            return np.zeros(shape, dtype=v.dtype)
        # calculate buffer spaces
        if self._batch is None:
            arr = next(iter(ub_utils.iter_nested(data)))
            assert not np.isscalar(arr), f'rank must be > 0'
            self._calc_space(len(arr))
        self._data = ub_utils.map_nested(data, op=_create_space)

    def _get_data(self, indices):
        assert not self.isnull, 'Buffer space not created'
        _slice_op = lambda v: v[indices]
        return ub_utils.map_nested(self._data, _slice_op)

    def _set_data(self, data, indices):
        assert indices is not None, '`indices` not set'
        # create space if the space is empty
        if self.isnull:
            self._melloc(data)
        # assign to buffer
        def _assign_op(data_tuple):
            new_data, data = data_tuple
            data[indices] = np.asarray(new_data).astype(data.dtype)
        ub_utils.map_nested_tuple((data, self._data), _assign_op)

    def load(self, path):
        raise NotImplementedError

    def save(self, path):
        raise NotImplementedError


class ReplayBuffer(BaseBuffer):
    '''A replay buffer to store nested type of data
    Note that you can only use python tuple and python
    dict to represent the nested data. A python list
    is treated as a single element.

    Example:
    >>> buffer = DictReplayBuffer(3, 1)
    >>> buffer.add(a=[0.1], b=([1], [True]))
    >>> buffer.add(a=[0.2], b=([2], [True]))
    >>> buffer.data
    {'a': array([[0.1, 0.2, 0.0]]),
        'b': (array([[1, 2, 0]]), array([[ True, True, False]]))}
    '''
    def keys(self):
        '''Return keys'''
        return [] if self.isnull else self._data.keys()
    
    def __contains__(self, key):
        '''Return True, if key is in the dict'''
        return not self.isnull and key in self._data.keys()

    def add(self, **data):
        return super().add(data)

    def update(self, indices, **data):
        return super().update(data, indices=indices)

    def _assert_keys_exist(self, keys):
        for key in keys:
            assert key in self, f'Key "{key}" does not exist'

    def _set_data(self, data, indices):
        if not self.isnull:
            self._assert_keys_exist(data.keys())
        return super()._set_data(data, indices)

class DynamicBuffer(ReplayBuffer):
    def __init__(self, batch: int=None):
        '''A dynamic replay buffer which has unlimited spaces to store 
        sequential nested data. This kind of buffer is used by on-policy 
        algo, e.g. PPO, which needs to reset buffer frequently and has an
        uncertained amount of samples. Note that `buffer.make()` must be 
        called before sampling data. This buffer has the following 
        procedure:
            __init__ -> add -> make -> update -> reset

        Args:
            batch (int, optional): batch size of replay samples, commonly 
                referred to as number of envs in a vectorized env. This 
                value is automatically set when user first adds a batch 
                of data. Defaults to None.
        '''
        self._batch = batch
        self._ready_for_sample = False
        self.reset()

    @property
    def capacity(self):
        '''Return the capacity of the buffer'''
        return float('inf')

    @property
    def slots(self):
        '''Return number of slots'''
        return self._pos

    @property
    def head(self):
        '''Return the index of the first slot'''
        return 0

    @property
    def ready_for_sample(self):
        '''Return True if buffer is ready for sampling'''
        return self._ready_for_sample

    def reset(self):
        '''Reset buffer'''
        super().reset()
        self._ready_for_sample = False

    def add(self, **data):
        '''Add new batch data into the buffer'''
        if self.ready_for_sample:
            raise RuntimeError('Buffer can\'t add data when it '
                'is ready for sampling')
        arr = next(iter(ub_utils.iter_nested(data)))
        assert not np.isscalar(arr), f'rank must be > 0'
        n_samples = len(arr)
        # copy data into the buffer
        cur_pos = self._pos
        self._append_data(data)
        # increase number of batch samples
        self._pos += 1
        return (cur_pos, np.arange(n_samples))

    def update(self, indices, **data):
        '''Update buffer contents
        You should call `make()` before calling `update()`
        '''
        if not self.ready_for_sample:
            raise RuntimeError('Call `buffer.make()` before calling '
                '`buffer.update()`')
        super().update(indices=indices, **data)

    def make(self):
        '''Prepare for sampling
        Convert list to np.ndarray
        '''
        if self.ready_for_sample:
            raise RuntimeError('The buffer has already made.')
        self._data = ub_utils.nested_to_numpy(self._data)
        self._ready_for_sample = True

    def len_slots(self):
        return self._pos

    def __len__(self):
        return 0 if self.isnull else self._pos*self._batch

    # --- private method ---
    def _calc_space(self, batch: int):
        '''Calculate buffer spaces'''
        assert isinstance(batch, int)
        assert batch > 0
        self._batch = batch

    def _melloc(self, data):
        '''Create buffer space'''
        _create_space = lambda v: []
        if self._batch is None:
            arr = next(iter(ub_utils.iter_nested(data)))
            assert not np.isscalar(arr), f'rank must be > 0'
            self._calc_space(len(arr))
        self._data = ub_utils.map_nested(data, op=_create_space)
    
    def _append_data(self, data):
        if self.isnull:
            self._melloc(data)
        if self.ready_for_sample:
            raise RuntimeError('The buffer can not append data after '
                'calling `buffer.make()`.')
        self._assert_keys_exist(data.keys())
        def _append_op(data_tuple):
            new_data, data = data_tuple
            data.append(np.asarray(new_data))
        ub_utils.map_nested_tuple((data, self.data), _append_op)

# === Sampler ===

class BaseSampler(metaclass=abc.ABCMeta):
    def __init__(self, buffer: BaseBuffer):
        if not isinstance(buffer, BaseBuffer):
            raise ValueError('`buffer` must be an instance of BaseBuffer, '
                f'got {type(buffer)}')
        self._buffer = buffer
        self._cached_inds = None
    
    @property
    def buffer(self):
        return self._buffer

    @property
    def indices(self):
        '''Cached indices'''
        return self._cached_inds

    @property
    def rel(self):
        '''Relative indexing, relative to sampled indices'''
        return RelativeIndex(
            self._buffer, 
            self._cached_inds,
            (self._buffer.len_slots(), self._buffer.batch)
        )

    def __call__(self, *args, **kwargs):
        '''A shortcut to self.sample()'''
        return self.sample(*args, **kwargs)

    @abc.abstractmethod
    def sample(self, batch_size, seq_len):
        '''Random sample replay buffer'''
        raise NotImplementedError

    def add(self, *args, **kwargs):
        '''Add samples to the buffer'''
        return self.buffer.add(*args, **kwargs)

    def update(self, *args, **kwargs):
        '''Update sampled data to the buffer'''
        assert self._cached_inds is not None
        return self.buffer.update(*args, **kwargs, indices=self._cached_inds)

class UniformSampler(BaseSampler):
    def __init__(self, buffer: BaseBuffer):
        '''Uniform sampler samples buffer uniformly.
        This sampler is mainly for off-policy algorithms.

        Args:
            buffer (BaseBuffer): Replay buffer
        '''
        super().__init__(buffer)

    def sample(self, batch_size: int=None, seq_len: int=None):
        '''Randomly sample a batch or a batch sequence of data from the 
        replay buffer

        Args:
            batch_size (int, optional): Batch size to sample. Defaults to None.
            seq_len (int, optional): Length of sequences. Defaults to None.
        
        Returns:
            Any: Returns a slice of data from the replay buffer. Each element has 
                shape (b, *data.shape) if seq_len is None, otherwise, 
                (b, seq, *data.shape).
        '''
        buf = self.buffer
        if not buf.ready_for_sample:
            raise RuntimeError('Buffer is not ready for sampling, '
                'call `buffer.make()` before sampling')
        if batch_size is None:
            batch_size = len(buf)
        if seq_len is None:
            inds = np.random.randint(len(buf), size=batch_size)
        else:
            inds = np.random.randint(len(buf)-seq_len*buf.batch, 
                                     size=(batch_size, 1))
        inds1, inds2 = buf.unravel_index(inds)
        if seq_len is not None:
            inds1 = inds1 + np.arange(seq_len)
            inds2 = np.broadcast_to(inds2, inds1.shape)
        # shift indices
        inds1 = (inds1 + buf.head) % buf.len_slots()
        self._cached_inds = (inds1, inds2)
        return self.buffer[self._cached_inds]

class PrioritizedSampler(BaseSampler):
    def __init__(self, buffer: ReplayBuffer,
                       alpha:      float, 
                       weight_key: str='w'):
        '''An implementation of Prioritized Experience Replay
        from "Prioritized Experience Replay".

        Args:
            buffer (ReplayBuffer): Replay buffer.
            alpha (float): Prioritization exponent.
            weight_key (str, optional): keys to store weights. One can access
                weights from the sampled batches with this key. Defaults to 'w'.
        '''
        super().__init__(buffer)
        if not isinstance(buffer, ReplayBuffer):
            raise ValueError(f'{type(self).__name__} does not support'
                f'{type(buffer).__name__}')
        if isinstance(buffer, DynamicBuffer):
            raise ValueError(f'{type(self).__name__} does not '
                'support DynamicBuffer')
        self._weight_tree = None
        self._weight_key  = weight_key
        self._alpha = alpha
        self._max_w = 1.0
        self._min_w = 1.0
        self._eps = np.finfo(np.float32).eps.item()

    def add(self, *args, **kwargs):
        '''Add samples to the buffer and weight tree'''
        _indices = self.buffer.add(*args, **kwargs)
        indices = self.buffer.ravel_index(_indices)
        if self._weight_tree is None:
            self._melloc()
        self._weight_tree[indices] = self._max_w**self._alpha
        return _indices

    def update(self, *args, **kwargs):
        '''Update sampled data to the buffer'''
        w = kwargs.pop(self._weight_key, None)
        if w is not None:
            # update tree weights
            w = np.abs(np.asarray(w)) + self._eps
            flat_w = w.flatten()
            inds = self.buffer.ravel_index(self._cached_inds)
            self._weight_tree[inds] = flat_w ** self._alpha
            self._max_w = max(self._max_w, np.max(w))
            self._min_w = min(self._min_w, np.min(w))
        return self.buffer.update(*args, **kwargs, indices=self._cached_inds)

    def sample(self, batch_size: int = None, 
                     seq_len:    int = None, 
                     beta:     float = 0.0):
        '''Randomly sample a batch or a batch sequence of data from 
        the replay buffer

        NOTE: The sequence sampling may sample a data that exceeds the 
        tail of the circular buffer (ReplayBuffer).

        Args:
            batch_size (int, optional): Batch size to sample. Defaults to None.
            seq_len (int, optional): Length of sequences. Defaults to None.
            beta (float, optional): Importance sampling exponent. Defaults to 0.0.

        Returns:
            Any: Returns a slice of data from the replay buffer. Each 
                element has shape (b, *data.shape) if seq_len is None, 
                otherwise, (b, seq, *data.shape).
        '''
        buf = self.buffer
        if not buf.ready_for_sample:
            raise RuntimeError('Buffer is not ready for sampling, '
                'call `buffer.make()` before sampling')
        if batch_size is None:
            batch_size = len(buf)
        if seq_len is None:
            samp = np.random.rand(batch_size) * self._weight_tree.sum()
        else:
            bound = self._weight_tree.sum()
            samp = np.random.rand(batch_size).reshape(-1, 1) * bound
        inds = self._weight_tree.index(samp)
        inds1, inds2 = buf.unravel_index(inds)
        if seq_len is not None:
            inds1 = inds1 + np.arange(seq_len)
            inds2 = np.broadcast_to(inds2, inds1.shape)
        inds1 = inds1 % buf.len_slots()
        self._cached_inds = (inds1, inds2)
        inds = buf.ravel_index(self._cached_inds)
        # calculate sample weights
        weight = (self._weight_tree[inds] / self._min_w) ** (-beta)
        weight = weight / np.max(weight)
        weight = weight.reshape(inds1.shape)
        # get batch data
        batch = self.buffer[self._cached_inds]
        batch[self._weight_key] = weight
        return batch

    # --- private methods ---
    def _melloc(self):
        '''Create segment tree space'''
        self._weight_tree = SegmentTree(self.buffer.capacity)

# alias
PriorSampler = PrioritizedSampler

class PermuteSampler(BaseSampler):
    def __init__(self, buffer: BaseBuffer):
        '''Permute sampler samples buffer by random permutation. It
        returns an iterator iterates through all rollout data.

        Args:
            buffer (BaseBuffer): Replay buffer.
        '''
        super().__init__(buffer)
    
    def sample(self, batch_size: int=None, seq_len: int=None):
        '''Create an iterator which randomly iterate through all data 
        in the replay buffer with a given batch size and sequence length.

        Args:
            batch_size (int, optional): Batch size to sample. Defaults to None.
            seq_len (int, optional): Length of sequences. If it's None, the sampled
                data has shape (b, *data.shape), otherwise, (b, seq, *data.shape).
                Defaults to None.
        
        Returns:
            Any: Returns an iterator which produces slices of data from the replay 
                buffer. Each element has shape (b, *data.shape) if seq_len is None, 
                otherwise, (b, seq, *data.shape).
        '''
        buf = self.buffer
        if not buf.ready_for_sample:
            raise RuntimeError('Buffer is not ready for sampling, '
                'call `buffer.make()` before sampling')
        return self._iter(batch_size=batch_size, seq_len=seq_len)

    def _iter(self, batch_size: int=None, seq_len: int=None):
        buf = self.buffer
        if batch_size is None:
            batch_size = len(buf)
        if seq_len is None:
            inds = np.arange(len(buf))
        else:
            #TODO: skip seq_len steps
            inds = np.arange(len(buf)-seq_len*buf.batch).reshape(-1, 1)
        inds1, inds2 = buf.unravel_index(inds)
        #inds1, inds2 = np.unravel_index(inds, (buf.slots, buf.batch))
        if seq_len is not None:
            inds1 = inds1 + np.arange(seq_len)
            inds2 = np.broadcast_to(inds2, inds1.shape)
        # shift indices
        inds1 = (inds1 + buf.head) % buf.len_slots()
        # shuffle the indices of the indices of the samples
        permute = np.arange(len(inds1))
        np.random.shuffle(permute)
        start_ind = 0
        while start_ind < len(inds):
            slice_ = range(start_ind, start_ind+batch_size)
            indices = np.take(permute, slice_, mode='wrap')
            self._cached_inds = (inds1[indices], inds2[indices])
            # return iterator
            yield self.buffer[self._cached_inds]
            start_ind += batch_size

# === data utils ===
def compute_nstep_rew(rew: np.ndarray,
                      done: np.ndarray,
                      gamma: float = 0.99):
    '''Compute N step rewards
    aggregate the folloing N-1 steps discounted rewards to the 
    first reward:
    rew = rew[0] + gamma**1 * rew[1] + gamma**2 * rew[2] + ...

    Args:
        rew (np.ndarray): Reward sequences, (step, ...)
        done (np.ndarray): Done sequences, (step, ...)
        gamma (float, optional): Discount factor. Defaults to 0.99

    Returns:
        np.ndarray: N step rewards (...)
    '''
    rew = np.asarray(rew).astype(np.float32)
    done = np.asarray(done).astype(np.float32)
    mask = 1.-done
    if len(done) == 0:
        return rew
    return _compute_nstep_rew(rew=rew, mask=mask, gamma=gamma)

def _compute_nstep_rew(rew: np.ndarray,
                       mask: np.ndarray,
                       gamma: float = 0.99):
    gams = gamma ** np.arange(len(mask))
    res = np.zeros_like(rew[0])
    prev_mask = np.ones_like(mask[0])
    for t in range(len(mask)):
        res += gams[t] * rew[t] * prev_mask
        prev_mask *= mask[t]
    return res

def compute_advantage(rew:        np.ndarray,
                      val:        np.ndarray,
                      done:       np.ndarray,
                      gamma:      float = 0.99,
                      gae_lambda: float = 1.0):
    '''Compute GAE

    Args:
        rewards (np.ndarray): Rewards (steps, ...)
        values (np.ndarray): Predicted values (steps, ...)
        dones (np.ndarray): Done flags (steps, ...)
        gamma (float, optional): Discount factor. Defaults to 0.99.
        gae_lambda (float, optional): GAE lambda. Defaults to 1.0.

    Returns:
        np.ndarray: GAE
    '''
    # ensure the inputs are np.ndarray
    rew  = np.asarray(rew).astype(np.float32)
    val  = np.asarray(val).astype(np.float32)
    done = np.asarray(done).astype(np.float32)
    mask = 1.-done
    return _compute_advantage(rew=rew, val=val, mask=mask,
        gamma=gamma, gae_lambda=gae_lambda)

def _compute_advantage(rew:        np.ndarray,
                       val:        np.ndarray,
                       mask:       np.ndarray,
                       gamma:      float,
                       gae_lambda: float):
    '''Perform GAE computation'''
    adv = np.zeros_like(rew)
    gae = 0.
    next_mask = mask[-1]
    next_val  = val[-1]
    for t in reversed(range(len(mask))):
        delta = rew[t] + gamma * next_val * next_mask - val[t]
        gae = delta + gamma * gae_lambda * next_mask * gae
        adv[t] = gae
        next_mask = mask[t]
        next_val  = val[t]
    return adv


# --- bulid in ---
import os
import sys
import time

from typing import Union

# --- 3rd party ---
import gym
import numpy as np

from gym.wrappers import TimeLimit

# --- my module ---
from unstable_baselines.lib import utils as ub_utils
from unstable_baselines.lib.envs import vec as ub_vec

__all__ = [
    'TimeLimit',
    'TimeFeatureWrapper',
    'wrap_mujoco',
    'ObsNorm'
]

# === Wrappers for Continuous Tasks ===

# Borrowed from Stable baselines
class TimeFeatureWrapper(gym.Wrapper):
    """
    Add remaining time to observation space for fixed length episodes.
    See https://arxiv.org/abs/1712.00378 and https://github.com/aravindr93/mjrl/issues/13.
    :param env: (gym.Env)
    :param max_steps: (int) Max number of steps of an episode
        if it is not wrapped in a TimeLimit object.
    :param test_mode: (bool) In test mode, the time feature is constant,
        equal to zero. This allow to check that the agent did not overfit this feature,
        learning a deterministic pre-defined sequence of actions.
    """
    def __init__(self, env, max_steps=1000, test_mode=True):
        super().__init__(env)
        assert isinstance(env.observation_space, gym.spaces.Box)

        low, high = env.observation_space.low, env.observation_space.high
        low, high = np.concatenate((low, [0.])), np.concatenate((high, [1.]))
        
        self.observation_space = gym.spaces.Box(low=low, high=high, dtype=np.float32)

        if (isinstance(env, TimeLimit)):
            self._max_steps = env._max_episode_steps
        else:
            self._max_steps = max_steps
        
        self._current_step = 0
        self._test_mode = test_mode

    def reset(self, **kwargs):
        self._current_step = 0
        obs = self.env.reset(**kwargs)
        return self._get_obs(obs)

    def step(self, action):
        self._current_step += 1
        obs, reward, done, info = self.env.step(action)
        return self._get_obs(obs), reward, done, info

    def _get_obs(self, obs):
        """
        Concatenate the time feature to the current observation.
        :param obs: (np.ndarray)
        :return: (np.ndarray)
        """
        # Remaining time is more general
        time_feature = 1 - (self._current_step / self._max_steps)
        if self._test_mode:
            time_feature = 1.0
        # Optionnaly: concatenate [time_feature, time_feature ** 2]
        return np.concatenate((obs, [time_feature])).astype(obs.dtype)


def wrap_mujoco(env:       gym.Env,
                time_limit:    int = 1000,
                time_feature: bool = True,
                test_mode:    bool = False):
    if time_limit:
        env = TimeLimit(env, time_limit)
    if time_feature:
        env = TimeFeatureWrapper(env, test_mode=test_mode)
    return env

# === Other wrappers ===

class ObsNorm(gym.Wrapper):
    def __init__(self,
        env: gym.Env,
        rms_norm: Union[str, bool, ub_utils.RMSNormalizer,
                        ub_vec.BaseVecEnv] = None,
        update_rms: bool = None,
    ):
        '''Observation normalization wrapper. Normalize observations with
        newly created running mean std or the existing RMSNormalizer from
        other vectorized environments. Note that if the RMSNormalizer is 
        disabled or fixed, then this wrapper is omitted.

        Args:
            env (gym.Emv): Environment to wrap.
            rms_norm (Union[str, ub_utils.RunningMeanStd, ub_vec.BaseVecEnv],
                optional): Running mean and std for normalizing observations.
                `str` is the path to the RunningMeanStd file. If it is a
                BaseVecEnv then this wrapper calls BaseVecEnv's `normalize` 
                method to perform normalizations. In this case, if BaseVecEnv's 
                `enable_norm` is disabled then this wrapper is also disabled. 
                Defaults to None.
            update_rms (bool, optional): This value is used to prevent from 
                updating others' rms normalizer. But you still can set this
                value to True if you want to. In default, if `rms` is a 
                BaseVecEnv, this value is set to False, otherwise to True.
        '''
        super().__init__(env)
        self.observation_space = env.observation_space
        self.action_space = env.action_space
        self._rms_norm = self._get_rms_opt(rms_norm)
        # prevent from updating others' rms normalizer
        self.update_rms = (not isinstance(rms_norm, ub_vec.BaseVecEnv)
            if update_rms is None else bool(update_rms))

    @property
    def rms_norm(self):
        return self._rms_norm

    def reset(self, **kwargs):
        obs = super().reset(**kwargs)
        if self.update_rms:
            self.rms_norm.update(obs)
        return self.rms_norm(obs)

    def step(self, act):
        obs, rew, done, info = super().step(act)
        if self.update_rms:
            self.rms_norm.update(obs)
        return self.rms_norm(obs), rew, done, info

    def load(self, path:str):
        '''load rms valu from path'''
        self.rms_norm.load(path)
        return self

    def save(self, path:str):
        self.rms_norm.save(path)

    def _get_rms_opt(self, opt):
        # using existing RMSNormalizer from vectorized environments
        if isinstance(opt, ub_vec.BaseVecEnv):
            opt = opt.rms_norm
        if isinstance(opt, ub_utils.RMSNormalizer):
            # setup RMS normalizer
            if not opt.is_setup:
                opt.setup(self.observation_space)
            return opt
        # select default RMS normalizer
        if isinstance(self.observation_space, 
                (gym.spaces.Dict, gym.spaces.Tuple)):
            # TODO _rms_class = ub_utils.NestedRMSNormalizer
            raise NotImplementedError
        else:
            _rms_class = ub_utils.RMSNormalizer
        enable = None
        if isinstance(opt, bool):
            enable = opt
        # Create RMSNormalizer
        rms = _rms_class(space  = self.observation_space,
                         enable = enable)
        if isinstance(opt, str):
            rms.load(opt)
        return rms


# --- built in ---
import os
import abc
import sys
import enum
import time
import functools

# --- 3rd party ---
import numpy as np

# --- my module ---
from unstable_baselines.lib import utils as ub_utils



__all__ = [
    'get_scheduler',
    'Scheduler',
    'ConstantScheduler',
    'ConstScheduler',
    'LinearScheduler',
    'ExponentialScheduler',
    'ExpScheduler',
    'MultiScheduler'
]

# === Units ===

@enum.unique
class Unit(str, enum.Enum):
    '''Note: the enum value must match
    to its name, because json.dumps will
    dump enum's value.
    '''
    timestep = 'timestep'
    epoch    = 'epoch'
    subepoch = 'subepoch'
    gradstep = 'gradstep'
    progress = 'progress'

# Registry class
class _RegisteredSchedulerClass(ub_utils.StateObject):
    '''Storing registered scheduler types'''
    # Default scheduler type
    default: 'Scheduler' = None
    def get_type(self, scheduler_type):
        '''Retrieve registered scheduler type by name or
        type

        Args:
            scheduler_type (str, type): scheduler type to retrieve

        Raises:
            ValueError: raised if `scheduler_type` is not contain in
                the registry.

        Returns:
            type: type of scheduler
        '''
        if (type(scheduler_type) is type and 
                issubclass(scheduler_type, Scheduler)):
            return scheduler_type
        elif scheduler_type in self.keys():
            return self[scheduler_type]
        else:
            raise ValueError('Got unknown scheduler type: {}'.format(scheduler_type))

_RegisteredScheduler = _RegisteredSchedulerClass()

def get_scheduler(*args, **kwargs):
    '''Create or retrieve schedulers
    a shortcut of Scheduler.get_scheduler
    '''
    return Scheduler.get_scheduler(*args, **kwargs)

def get_scheduler_type(typename):
    '''Retrieve scheduler types by scheduler name or type
    A shortcut of _RegisteredScheduler.get_type
    '''
    return _RegisteredScheduler.get_type(typename)

# === Schedulers helpers ====

def _get_unit_enum(item):
    if isinstance(item, Unit):
        return item
    if isinstance(item, str):
        return Unit(item) # may raises KeyError
    raise ValueError(f'Unknown type of unit: {type(item)}')

def _get_state_key(unit: Unit):
    '''Retrieve the state key corresponding to the unit'''
    if unit is Unit.timestep:
        key = 'num_timesteps'
    elif unit is Unit.epoch:
        key = 'num_epochs'
    elif unit is Unit.subepoch:
        key = 'num_subepochs'
    elif unit is Unit.gradstep:
        key = 'num_gradsteps'
    elif unit is Unit.progress:
        key = 'progress'
    else:
        #TODO: custom units
        # first check if unit is a Unit type. If not, 
        # check if it is a custom units.
        all_units = list(key.name for key in Unit)
        raise RuntimeError('Unknown unit, unit must be either of '
            f'{all_units}, got {unit}')
    return key

def _check_state_and_unit(unit: Unit, state: ub_utils.StateObject):
    '''Check if the state key corresponds to unit does exist
    '''
    if state is None:
        raise ValueError('State is empty, call '
            'bind() to set state.')
    if not isinstance(state, ub_utils.StateObject):
        raise TypeError('State must be a StateObject, '
            f'got {type(state)}')
    state_key = _get_state_key(unit)
    if state_key not in state.keys():
        raise KeyError(f'State does not contain key for {unit}')

def _get_steps_from_state(key: str, state: ub_utils.StateObject):
    '''Retrieve state info by state key
    Note that is function does not handle the exception of
    missing key, so ensure you have called `_check_state_and_unit`
    before calling this function.
    '''
    if state is None:
        raise ValueError('State is empty, call '
            'bind() to set state.')
    return state[key] # raise KeyError

# === Schedulers ===

class Scheduler(ub_utils.StateObject):
    '''Base scheduler class
    To create a new scheduler class, you need to implement the 
    following methods:
    * calc(self, steps):
    '''
    # do not serialize these members
    __slots__ = ['state', 'state_key']
    
    def __new__(cls, *args, **kwargs):
        sche = super().__new__(cls, *args, **kwargs)
        sche.type = cls.__name__

        return sche

    def __init__(self, unit: str='timestep', 
                       state: ub_utils.StateObject=None):
        '''Initialize scheduler

        Args:
            unit (str, optional): Unit. Defaults to 'timestep'.
            state (StateObject, optional): state object. Defaults to None.
        '''
        self.unit = _get_unit_enum(unit)
        self.state_key = _get_state_key(self.unit) # not serializable
        self.state = None # not serializable

        if state is not None:
            self.bind(state)

    def bind(self, state: ub_utils.StateObject):
        '''Bind state object'''
        _check_state_and_unit(self.unit, state)
        self.state = state

    @abc.abstractmethod
    def calc(self, steps):
        '''Calculate variables

        Args:
            steps (int, float): timestep, epoch, gradstep or progress

        Returns:
            float: interpolated value
        '''
        raise NotImplementedError

    def get_steps(self):
        if self.state is None:
            raise RuntimeError('The `state` is empty, call '
                                    'bind() to set `state`.')
        return _get_steps_from_state(self.state_key, self.state)

    def __call__(self, steps=None):
        
        if steps is None:
            steps = self.get_steps()

        return self.calc(steps)

    @classmethod
    def register(cls, names=[], default=False):
        '''Register a new type of Scheduler

        Args:
            names (list, optional): Name alias. Defaults to [].
            default (bool, optional): Set as default scheduler. 
                Defaults to False.
        '''
        def _register(scheduler_type):
            name_list = names
            if not isinstance(name_list, (tuple, list)):
                name_list = [name_list]
            for name in name_list:
                if name in _RegisteredScheduler.keys():
                    raise ValueError(f'The name {name} is used by scheduler'
                        f'{type(_RegisteredScheduler[name])}')

                _RegisteredScheduler[name] = scheduler_type
            if hasattr(scheduler_type, '__name__'):
                name = scheduler_type.__name__
                _RegisteredScheduler[name] = scheduler_type
            # set default scheduler
            if default:
                _RegisteredScheduler.default = scheduler_type
            return scheduler_type
        return _register

    @classmethod
    def get_scheduler(cls, *args, type=None, state=None, **kwargs):
        '''Get scheduler
        The first argument can be a defined Scheduler or a serialized
        Scheduler (mostly it's a `dict`). This is for restoring scheduler.
        Or you can specify `type` to create a new scheduler.

        Args:
            type (str, type, optional): Scheduler name or type. 
                Defaults to None.
            state (StateObject, optional): state object to bind. 
                Defaults to None.

        Returns:
            Scheduler: scheduler
        '''
        sche = None
        # Restore scheduler
        if len(args) == 1:
            if isinstance(args[0], Scheduler):
                # Scheduler(LinearScheduler(...))
                sche = args[0]
            elif isinstance(args[0], dict):
                # Scheduler({'type': 'linear', 'start_value': 1e-3...})
                params = args[0]
                params.update(kwargs)
                sche = cls.get_scheduler(**params)
        # Create a new scheduler
        if sche is None:
            if type is None:
                type = _RegisteredScheduler.default
            # get scheduler class
            type = _RegisteredScheduler.get_type(type)
            # create scheduler
            sche = type(*args, **kwargs)
        # bind state object
        if sche is not None:
            if isinstance(state, ub_utils.StateObject):
                sche.bind(state)
            return sche
        raise ValueError('Unknown scheduler params:'
            f'args={args}, type={type}, kwargs={kwargs}')

# === Build-in schedulers ===

@Scheduler.register(['const', 'constant',
                    'ConstScheduler',
                    'Constant'], default=True)
class ConstantScheduler(Scheduler):
    def __init__(self, value: float, **kwargs):
        '''Constant scheduler

        Args:
            value (float): variable values.
        '''
        super().__init__(**kwargs)
        self.value = value

    def __call__(self, steps=None):
        return self.value

# alias
ConstScheduler = ConstantScheduler
Constant = ConstantScheduler

@Scheduler.register(['linear', 'Linear'])
class LinearScheduler(Scheduler):
    def __init__(self, start_value: float,
                       decay_steps: int,
                       stop_value: float=0.0,
                       **kwargs):
        '''LinearScheduler
        Perform linear decay

        Args:
            start_value (float): Initial value.
            decay_steps (int, float): total steps to stop decay.
            stop_value (float, optional): minimum value. Default to 0.0
        '''
        super().__init__(**kwargs)

        self.start_value = start_value
        self.decay_steps = decay_steps
        self.stop_value  = stop_value

    def calc(self, steps):
        progress = min(1., float(steps)/float(self.decay_steps))
        return self.start_value + (self.stop_value - self.start_value) * progress


# alias
Linear = LinearScheduler

@Scheduler.register(['exp', 'ExpScheduler',
                    'Exponential'])
class ExponentialScheduler(Scheduler):
    def __init__(self, start_value: float, 
                       decay_steps: int, 
                       decay_rate: float, 
                       stop_value: float=None, 
                       **kwargs):
        '''ExponentialScheduler
        Perform exponential decay

        Args: TODO
            start_value (float): [description]
            decay_steps (int): [description]
            decay_rate (float): [description]
            stop_value (float, optional): minimum value. Defaults to None.
        '''        
        super().__init__(**kwargs)

        self.start_value = start_value
        self.decay_steps = decay_steps
        self.decay_rate = decay_rate
        self.stop_value = stop_value

    def calc(self, steps):
        decay = self.decay_rate**(float(steps)/self.decay_steps)
        v = self.start_value * decay
        if self.stop_value is not None:
            return np.maximum(self.stop_value, v)
        else:
            return v
            

# alias
ExpScheduler = ExponentialScheduler
Exponential = ExponentialScheduler

@Scheduler.register(['multi'])
class MultiScheduler(Scheduler):
    __slots__ = ['_op_func']
    def __init__(self, schedulers, op='min', **kwargs):
        '''Multiple schedulers
        This scheduler allows you to combine multiple schedulers.
        Note that the serializable operations `op` only supports
        'min', 'max' and 'mean'. If you are using custom operations,
        don't forget to provide them on restoring schedulers.
        Note that `unit` is discarded.

        Args:
            schedulers (list): A list of schedulers
            op (str, Callable, optional): operation to calculate
                final values. Defaults to 'min'.
        '''        
        super().__init__(**kwargs)
        self.schedulers = [
            Scheduler.get_scheduler(sche)
            for sche in schedulers
        ]
        assert len(self.schedulers) > 0
        if isinstance(op, str): # min, max, mean
            op = op.lower()
            self.op = op
            self._op_func = self.get_op_func(op)
        elif callable(op): # custom function
            self.op = None
            self._op_func = op
        else:
            raise ValueError(f'`op` must be str or callable, got {type(op)}')

    def get_op_func(self, op: str):
        assert op in ['max', 'min', 'mean']
        self.op = op
        if op == 'max':
            _op_func = np.max
        elif op == 'min':
            _op_func = np.min
        elif op == 'mean':
            _op_func = np.mean
        return functools.partial(_op_func, axis=0)

    def bind(self, state):
        super().bind(state)
        for sche in self.schedulers:
            sche.bind(state)

    def __call__(self, steps=None):
        return self.calc(steps)

    def calc(self, steps):
        vals = [sche(steps) for sche in self.schedulers]
        return self._op_func(vals)


# --- bulid in ---
import os
import abc
import cv2
import time

from typing import Union
from collections import deque

# --- 3rd party ---
import gym
import numpy as np
import tensorflow as tf

# --- my module ---

__all__ = [
    'NoopResetEnv',
    'MaxAndSkipEnv',
    'EpisodicLifeEnv',
    'FireResetEnv',
    'WarpFrame',
    'ScaledFloatFrame',
    'ClipRewardEnv',
    'FrameStack',
    'make_atari',
    'wrap_deepmind',
]

# === Atari Wrappers ===
'''
Mostly borrowed from OpenAI baselines and Stable baselines.
- Stable baselines: https://github.com/hill-a/stable-baselines
- OpenAI baselines: https://github.com/openai/baselines
'''

class NoopResetEnv(gym.Wrapper):
    def __init__(self, env, noop_max=30):
        """Sample initial states by taking random number of no-ops on reset.
        No-op is assumed to be action 0.
        """
        super().__init__(env)
        self.noop_max = noop_max
        self.noop_action = 0
        assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

    def reset(self, **kwargs):
        """ Do no-op action for a number of steps in [1, noop_max]."""
        self.env.reset(**kwargs)
        noops = self.unwrapped.np_random.randint(1, self.noop_max + 1)
        obs = None
        for _ in range(noops):
            obs, _, done, _ = self.env.step(self.noop_action)
            if done:
                obs = self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)

class MaxAndSkipEnv(gym.Wrapper):
    def __init__(self, env, skip=4):
        """Return only every `skip`-th frame"""
        super().__init__(env)
        # most recent raw observations (for max pooling across time steps)
        self._skip = skip

    def step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        total_reward = 0.0
        done = False
        obs_list = []
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            obs_list.append(obs)
            total_reward += reward
            if done:
                break
        # Note that the observation on the done=True frame
        # doesn't matter
        max_frame = np.max(obs_list[-2:], axis=0)
        return max_frame, total_reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)

class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        super().__init__(env)
        self.lives = 0
        self.was_real_done  = True

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert sometimes we stay in lives == 0 condition for a few frames
            # so it's important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        self.lives = lives
        return obs, reward, done, info

    def reset(self, **kwargs):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            obs = self.env.reset(**kwargs)
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
        self.lives = self.env.unwrapped.ale.lives()
        return obs

class FireResetEnv(gym.Wrapper):
    def __init__(self, env):
        """Take action on reset for environments that are fixed until firing."""
        super().__init__(env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def reset(self, **kwargs):
        self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)

# from Stable baselines
class WarpFrame(gym.ObservationWrapper):
    def __init__(self, env, width=84, height=84):
        """
        Warp frames to 84x84 as done in the Nature paper and later work.
        :param env: (Gym Environment) the environment
        """
        super().__init__(env)
        self.width = width
        self.height = height
        self.observation_space = gym.spaces.Box(low=0, high=255, 
                                            shape=(self.height, self.width, 1),
                                            dtype=env.observation_space.dtype)

    def observation(self, frame):
        """
        returns the current observation from a frame
        :param frame: ([int] or [float]) environment frame
        :return: ([int] or [float]) the observation
        """
        frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        frame = cv2.resize(frame, (self.width, self.height), interpolation=cv2.INTER_AREA)
        return frame[:, :, None]

class ScaledFloatFrame(gym.ObservationWrapper):
    def __init__(self, env):
        super().__init__(env)
        low = np.min(env.observation_space.low).astype(np.float32)
        high = np.max(env.observation_space.high).astype(np.float32)
        self.bias = low
        self.scale = high - low
        self.observation_space = gym.spaces.Box(
            low=0., high=1., shape=env.observation_space.shape,
            dtype=np.float32)

    def observation(self, observation):
        obs = np.asarray(observation).astype(np.float32)
        return (obs - self.bias) / self.scale

class ClipRewardEnv(gym.RewardWrapper):
    reward_range = (-1, 1)
    def __init__(self, env):
        super().__init__(env)

    def reward(self, reward):
        """Bin reward to {+1, 0, -1} by its sign."""
        return np.sign(reward)

class FrameStack(gym.Wrapper):
    def __init__(self, env, n_frames=4):
        super().__init__(env)
        self.n_frames = n_frames
        self.frames = deque([], maxlen=n_frames)
        space = env.observation_space
        low = np.min(space.low)
        high = np.max(space.high)
        # stack to the last dim (channel)
        shape = (*space.shape[:-1], space.shape[-1] * n_frames)
        self.observation_space = gym.spaces.Box(
            low=low, high=high, 
            shape=shape,
            dtype=env.observation_space.dtype
        )
    
    def reset(self, **kwargs):
        obs = self.env.reset(**kwargs)
        for _ in range(self.n_frames):
            self.frames.append(obs)
        return self._get_ob()

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.frames.append(obs)
        return self._get_ob(), reward, done, info

    def _get_ob(self):
        assert len(self.frames) == self.n_frames
        return np.concatenate(self.frames, axis=-1)


def make_atari(env_id):
    '''Make atari environment, random start + frameskip'''
    env = gym.make(env_id)
    assert 'NoFrameskip' in env.unwrapped.spec.id
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    return env

def wrap_deepmind(env:       gym.Env,
                  episode_life: bool = True, 
                  clip_rewards: bool = True,
                  frame_stack:   int = 4, 
                  scale:        bool = False, 
                  warp_frame:   bool = True):
    """Configure environment for DeepMind-style Atari. The observation is
    channel-first: (c, h, w) instead of (h, w, c).
    :param str env_id: the atari environment id.
    :param bool episode_life: wrap the episode life wrapper.
    :param bool clip_rewards: wrap the reward clipping wrapper.
    :param int frame_stack: wrap the frame stacking wrapper.
    :param bool scale: wrap the scaling observation wrapper.
    :param bool warp_frame: wrap the grayscale + resize observation wrapper.
    :return: the wrapped atari environment.
    """
    assert 'NoFrameskip' in env.unwrapped.spec.id
    if episode_life:
        env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    if warp_frame:
        env = WarpFrame(env)
    if scale:
        env = ScaledFloatFrame(env)
    if clip_rewards:
        env = ClipRewardEnv(env)
    if frame_stack:
        env = FrameStack(env, frame_stack)
    return env



# --- built in ---
import os
import sys
import time
import logging
import datetime

# --- 3rd party ---
import gym
import numpy as np
import tensorflow as tf

# --- my module ---
import unstable_baselines as ub

class QNet(ub.nets.ValueNet):
    def __init__(self, action_dims: int,
                       dueling: bool = False,
                       net: tf.keras.Model = None,
                       **kwargs):
        '''Q-value net

        Args:
            action_dims (int): Output action size.
            dueling (bool, optional): Whether to use dueling nets. 
                Defaults to False.
            net (tf.keras.Model, optional): Base network, feature extractor. 
                Defaults to None.
        '''
        super().__init__(net=net, **kwargs)
        self.action_dims = action_dims
        self.dueling = dueling

    def call(self, inputs, training=True):
        '''Forward q-value net

        Args:
            inputs (tf.Tensor): Input tensors, shape (b, *)
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: value predictions, shape (b, act)
        '''
        x = self._model(inputs, training=training)
        if self.dueling:
            # Dueling DQN
            q = self._value[0](x, training=training)
            v = self._value[1](x, training=training)
            a = q - tf.math.reduce_mean(q, axis=-1, keepdims=True)
            return a + v
        else:
            return self._value(x, training=training)

    def create_value_model(self):
        '''Create value model
        Override this method to customize model arch

        Returns:
            tf.keras.Model, list: Q net. If `dueling`, returns
                a list contains Q net and V net.
        '''
        if self.dueling:
            # Dueling DQN (Q, V)
            return [
                tf.keras.layers.Dense(self.action_dims),
                tf.keras.layers.Dense(1)
            ]
        else:
            return tf.keras.layers.Dense(self.action_dims)

class Agent(ub.base.BaseAgent):
    support_obs_spaces = ub.utils.spaces.All
    support_act_spaces = [gym.spaces.Discrete]
    def __init__(self, observation_space,
                       action_space,
                       dueling:   bool = False,
                       force_mlp: bool = False,
                       mlp_units: list = [64, 64],
                       **kwargs):
        '''DQN agent

        The `observation_space` and `action_space` can be `None` for delayed
        model setup. You should call `set_spaces()` then call `setup()` to 
        manually setup the model.

        Args:
            observation_space (gym.Space): The observation space of the env.
                Set to None for delayed setup.
            action_space (gym.Space): The action space of the env. Set to None 
                for delayed setup.
            force_mlp (bool, optional): Force to use mlp as feature extractors.
                Default to False.
            mlp_units (list, optional): Number of units for each mlp layer. 
                Defaults to [64, 64]
        '''        
        super().__init__(observation_space, action_space, **kwargs)

        self.dueling   = dueling
        self.force_mlp = force_mlp
        self.mlp_units = mlp_units

        self.value = None

        if (observation_space is not None) and (action_space is not None):
            self.setup()
    
    # --- setup group ---
    def _setup_model(self):
        '''Setup models
        Override this method to customize nets
        '''
        # AwesomeNet automatically handles complex observations
        # e.g. Dict or Tuple and creates separate feature extractors
        # either NatureCnn or MlpNet for each space
        base_net = ub.nets.AwesomeNet(
            input_space = self.observation_space,
            force_mlp   = self.force_mlp,
            mlp_units   = self.mlp_units
        )
        # create q net
        self.value = QNet(
            action_dims = self.action_space.n,
            dueling     = self.dueling,
            net         = base_net
        )

    def setup(self):
        '''(Override) Setup agent model'''
        self._setup_model()
        # construct network params
        inputs = ub.utils.input_tensor(self.observation_space)
        self.call(inputs, proc_obs=True)

    def call_value(self, inputs, proc_obs=True, training=True):
        '''Forward value net'''
        if proc_obs:
            inputs = self.proc_observation(inputs)
        values = self.value(inputs, training=training)
        return values

    @tf.function
    def call(self, inputs,
                   proc_obs: bool = True,
                   training: bool = True):
        '''(Override) Forward agent

        Args:
            inputs (tf.Tensor): Batch observations in shape
                (b, *obs.space.shape), tf.float32
            proc_obs (bool, optional): Preprocess observations. Defaults to True.
            training (bool, optional): Training mode. Defaults to True.

        Returns:
            tf.Tensor: Predicted actions, shape (b,)
            tf.Tensor: Predicted q values, shape (b,)
        '''
        # forward value nets
        values  = self.call_value(inputs, proc_obs, training)
        actions = tf.math.argmax(values, axis=-1)
        values  = tf.math.reduce_max(values, axis=-1)
        return actions, values

    def predict(self, inputs, proc_obs: bool=True):
        return super().predict(inputs, proc_obs=proc_obs)

    def get_config(self):
        config = super().get_config()
        config.update({
            'dueling':   self.dueling,
            'force_mlp': self.force_mlp,
            'mlp_units': self.mlp_units
        })
        return config

class DQN(ub.base.OffPolicyModel):
    support_obs_spaces = ub.utils.spaces.All
    support_act_spaces = [gym.spaces.Discrete]
    def __init__(self,
        env,
    # --- hyper parameters ---
        learning_rate: float = 3e-4,
        buffer_size:     int = int(1e5),
        multi_step:      int = 1,
        gamma:         float = 0.99,
        tau:           float = 1.0,
        reg_coef:      float = 0.0,
        clipnorm:      float = 0.5,
        explore_rate:  float = 0.3,
        huber:          bool = True,
        huber_rate:    float = 0.1,
        prioritized:    bool = False,
        prio_alpha:    float = 0.6,
        prio_beta:     float = 0.4,
    # --- architecture parameters ---
        dueling:        bool = False,
        force_mlp:      bool = False,
        mlp_units:      list = [64, 64],
    # --- other parameters ---
        n_steps:         int = 4,
        n_gradsteps:     int = 1,
        warmup_steps:    int = int(1e4),
        batch_size:      int = 128,
        **kwargs
    ):
        '''DQN
        Variants: Double / Dualing / N-step DQN

        The implementation mainly follows its originated paper
        `Deep Reinforcement Learning with Double Q-learning` by Hasselt et al.
        `Dueling Network Architectures for Deep Reinforcement Learning` 
        by Wang et al.

        The first argument `env` can be `None` for delayed model setup. You
        should call `set_env()` then call `setup_model()` to manually setup
        the model.

        Args:
            env (gym.Env): Training environment. Can be `None`
            learning_rate (float, optional): Learning rate can be a float or
                a ub.sche.Scheduler. Defaults to 3e-4.
            buffer_size (int, optional): Replay buffer capacity. Defaults to 
                int(1e6).
            multi_step (int, optional): N-step learning. Defaults to 1.
            gamma (float, optional): Discount factor. Defaults to 0.99.
            tau (float, optional): Soft update parameter. Defaults to 1.0.
            reg_coef (float, optional): Regularization coefficient. Defaults to 0.0.
            clipnorm (float, optional): Gradient clip. Defaults to 0.5.
            huber (bool, optional): Whether to use Huber loss or MSE loss. Defaults to True.
            huber_rate (float, optional): Rate of quadratic and linear in huber loss.
                Defaults to 0.1.
            explore_rate (float, optional): Random exploration rate (epsiln greedy).
                It can be a float for constant exploration or a ub.sche.Scheduler. 
                Defaults to 0.3.
            prioritized (bool, optional): Whether to use prioritized replay buffer.
                Defaults to False.
            prio_alpha (float, optional): Prioritization exponent. Defaults to 0.6.
            prio_beta: (float, optional): Importance sampling exponent, can be a float
                or a ub.sche.Scheduler. Defaults to 0.4.
            dueling (bool, optional): Whether to enable dueling DQN. Defaults to False.
            force_mlp (bool, optional): Force to use MlpNet as the feature extractors. Defaults to False.
            mlp_units (list, optional): Number of units for each mlp layer. Defaults to [64, 64].
            n_steps (int, optional): Number of steps to collect rollouts in
                each epoch. Defaults to 4.
            n_gradsteps (int, optional): Number of gradients in each epoch. Defaults to 1.
            warmup_steps (int, optional): Number of steps in warming up stage. Defaults to int(1e4).
            batch_size (int, optional): Training batch size. Defaults to 128.
        '''        
        super().__init__(
            env          = env,
            n_steps      = n_steps,
            n_gradsteps  = n_gradsteps,
            warmup_steps = warmup_steps,
            batch_size   = batch_size,
            **kwargs
        )
        self.learning_rate = learning_rate
        self.buffer_size   = buffer_size
        self.multi_step    = multi_step
        self.gamma         = gamma
        self.tau           = tau
        self.reg_coef      = reg_coef
        self.clipnorm      = clipnorm
        self.explore_rate  = explore_rate
        self.huber         = huber
        self.huber_rate    = huber_rate
        self.prioritized   = prioritized
        self.prio_alpha    = prio_alpha
        self.prio_beta     = prio_beta
        self.dueling       = dueling
        self.force_mlp     = force_mlp
        self.mlp_units     = mlp_units
        # initialize objects
        self.buffer    = None
        self.agent     = None
        self.agent_tar = None
        self.optimizer = None

        if env is not None:
            self.setup()

    # --- setup group ---
    def _setup_buffer(self):
        '''Setup buffers
        Override this method to customize to your replay buffer
        '''
        self.buffer  = ub.data.ReplayBuffer(self.buffer_size)
        if self.prioritized:
            self.sampler = ub.data.PriorSampler(self.buffer, self.prio_alpha)
        else:
            self.sampler = ub.data.UniformSampler(self.buffer)

    def _setup_model(self):
        '''Setup models, agent, target agent
        Override this method to customize to your model
        '''
        self.agent = Agent(
            self.observation_space,
            self.action_space,
            dueling   = self.dueling,
            force_mlp = self.force_mlp,
            mlp_units = self.mlp_units
        )
        self.agent_tar = Agent(
            self.observation_space,
            self.action_space,
            dueling   = self.dueling,
            force_mlp = self.force_mlp,
            mlp_units = self.mlp_units
        )
        self.agent_tar.update(self.agent, polyak=1.0)

    def _setup_optimizer(self):
        '''Setup optimizers
        Override this method to customize to your optimizer
        '''
        # --- setup scheduler ---
        self.learning_rate = ub.sche.get_scheduler(
            self.learning_rate,
            state = self.state
        )
        self.explore_rate = ub.sche.get_scheduler(
            self.explore_rate,
            state = self.state
        )
        if self.prioritized:
            self.prio_beta = ub.sche.get_scheduler(
                self.prio_beta,
                state = self.state
            )
        # --- optimizers ---
        self.optimizer = tf.keras.optimizers.Adam(
            learning_rate = self.learning_rate(),
            clipnorm = self.clipnorm
        )
        # track weights
        ub.utils.set_optimizer_params(
            self.optimizer,
            self.agent.trainable_variables
        )

    def setup(self):
        '''(Override) Setup models, optimizers and schedulers for training'''
        self._setup_buffer()
        self._setup_model()
        self._setup_optimizer()

    # --- forward group ---
    @tf.function
    def call(self, inputs,
                   proc_obs: bool = True,
                   training: bool = True):
        '''(Override) See Agent.call'''
        return self.agent(
            inputs,
            proc_obs = proc_obs,
            training = training
        )

    def predict(self, inputs,
                      proc_obs: bool = True):
        '''(Override) See Agent.predict'''
        return self.agent.predict(
            inputs,
            proc_obs = proc_obs
        )

    # --- collect group ---
    def _collect_step(self, obs):
        '''(Override) Collect one sample
        
        Args:
            obs (np.ndarray, optional): Current observations
        
        Returns:
            np.ndarray: next observations
        '''
        if (self.is_warming_up() or
                np.random.rand() < self.explore_rate()):
            # sample random action
            act = np.asarray([s.sample() for s in self.env.action_spaces])
        else:
            act = self.predict(obs)
        # step environment
        next_obs, rew, done, infos = self.env.step(act)
        # Add data to buffer
        self.sampler.add(
            obs      = obs,
            act      = act,
            done     = done,
            rew      = rew
        )
        return next_obs

    # --- train group ---
    def td_error(self, obs, act, done, rew, next_obs):
        '''Compute N-step TD-error

        Y = r + gamma*N * max(Q(s'))
        td = Y - Q(s)

        Args:
            obs (tf.Tensor): Batch observations, (b, *obs_space.shape)
            act (tf.Tensor): Batch actions, (b,), tf.int64
            done (tf.Tensor): Batch dones, (b,), tf.float32
            rew (tf.Tensor): Batch rewards, (b,), tf.float32
            next_obs (tf.Tensor): Batch next observations, (b, *obs_space.shape)
        
        Returns:
            tf.Tensor: TD error, tf.float32
        '''
        act  = tf.cast(act, dtype=tf.int64)
        rew  = tf.cast(rew, dtype=tf.float32)
        done = tf.cast(done, dtype=tf.float32)
        # calculate target q value
        next_qs = self.agent_tar.call_value(next_obs)
        next_q  = tf.math.reduce_max(next_qs, axis=-1)
        gamma   = self.gamma ** self.multi_step
        y = rew + tf.stop_gradient((1.-done) * gamma * next_q)
        # calculate current q
        qs = self.agent.call_value(obs)
        q  = tf.gather(qs, indices=act, batch_dims=1)
        return y - q

    def td_loss(self, td: tf.Tensor, w: tf.Tensor=None):
        '''Compute TD loss (MSE or Huber)from TD error

        Huber:
            1/2 * TD^2, if |TD| <= delta
            delta * (|TD| - delta/2), otherwise
        MSE:
            TD^2

        Args:
            td (tf.Tensor): TD error, tf.float32
            w (tf.Tensor, optional): Loss weights, tf.float32. Defaults to None.
        
        Returns:
            tf.Tensor: td loss, tf.float32
        '''
        if w is None:
            w = tf.ones_like(td)
        w = tf.cast(w, dtype=tf.float32)
        if self.huber:
            # huber loss
            delta = tf.convert_to_tensor(self.huber_rate)
            losses = tf.where(tf.math.abs(td) <= delta, 
                0.5 * tf.math.square(td),
                delta * tf.math.abs(td) - 0.5 * tf.math.square(delta))
        else:
            # mse loss
            losses = tf.math.square(td)
        return tf.math.reduce_mean(losses * w)

    def reg_loss(self, var_list):
        '''L2 regularization loss

        Args:
            var_list (list): A list of variables need to be regularized.

        Returns:
            tf.Tensor: regularization loss.
        '''
        reg = tf.keras.regularizers.L2(self.reg_coef)
        return tf.math.add_n([reg(var) for var in var_list] + [0.0])

    @tf.function
    def _train_model(self, batch, w=None):
        '''Perform one gradient update
        
        Args:
            batch (dict): A sample of batch.
            w (tf.Tensor, optional): Loss weights.
        '''
        vars = self.agent.trainable_variables
        with tf.GradientTape() as tape:
            tape.watch(vars)
            td = self.td_error(**batch)
            # compute losses
            td_loss  = self.td_loss(td, w=w)
            reg_loss = self.reg_loss(vars)
            # aggregate losses
            loss = (td_loss + reg_loss)
        # perform gradients
        grads = tape.gradient(loss, vars)
        self.optimizer.apply_gradients(zip(grads, vars))
        return {
            'loss':     loss,
            'td_loss':  td_loss,
            'reg_loss': reg_loss
        }, tf.math.abs(td)

    def _sample_nstep_batch(self, batch_size):
        '''Sample a batch of replay with n-step rewards (if enabled)
        from replay buffer or prioritized replay buffer
        '''
        if self.prioritized:
            beta = self.prio_beta()
            batch = self.sampler(batch_size, beta=beta)
        else:
            batch = self.sampler(batch_size)
        # get nstep rewards and dones
        if self.multi_step > 1:
            n_batch = self.sampler.rel[:self.multi_step]
            # calculate N-step rewards
            batch['rew'] = ub.data.compute_nstep_rew(
                rew   = n_batch['rew'],
                done  = n_batch['done'],
                gamma = self.gamma
            )
            batch['done'] = np.any(n_batch['done'], axis=0)
        # Get the observation after N steps
        batch['next_obs'] = self.sampler.rel[self.multi_step]['obs']
        return batch

    def _train_step(self, batch_size):
        '''(Override) Train one step'''
        # update learning rate
        lr = tf.convert_to_tensor(self.learning_rate())
        self.optimizer.lr.assign(lr)
        # sample data and train model
        batch = self._sample_nstep_batch(batch_size)
        losses, td = self._train_model(batch, batch.pop('w', None))
        if self.prioritized:
            # update sample weights
            self.sampler.update(w=td)
        return losses

    def _update_target(self):
        '''(Override) Update target networks'''
        self.agent_tar.update(self.agent, polyak=self.tau)

    def get_config(self):
        config = super().get_config()
        config.update({
            'learning_rate': self.learning_rate,
            'buffer_size':   self.buffer_size,
            'multi_step':    self.multi_step,
            'gamma':         self.gamma,
            'tau':           self.tau,
            'reg_coef':      self.reg_coef,
            'clipnorm':      self.clipnorm,
            'explore_rate':  self.explore_rate,
            'huber':         self.huber,
            'huber_rate':    self.huber_rate,
            'prioritized':   self.prioritized,
            'prio_alpha':    self.prio_alpha,
            'prio_beta':     self.prio_beta,
            'dueling':       self.dueling,
            'force_mlp':     self.force_mlp,
            'mlp_units':     self.mlp_units,
        })
        return config


import os
import sys
import time
import logging

class Color(object):

    # === Set ===
    _SET_ = set([1, 2, 4, 5, 7, 8])
    _RESET_ = set([0, 21, 22, 24, 25, 27, 28, 39, 49])

    # === Color ===
    _FDCOLOR_ = set([30, 31, 32, 33, 34, 35, 36, 37])
    _FLCOLOR_ = set([90, 91, 92, 93, 94, 95, 96, 97])
    _BDCOLOR_ = set([40, 41, 42, 43, 44, 45, 46, 47])
    _BLCOLOR_ = set([100, 101, 102, 103, 104, 105, 106, 107])

    def __init__(self, *args):
        self.args = [int(i) for i in args]

    def __int__(self):
        if len(self.args) > 1:
            print(self.args)
            raise Exception('too many color codes, only one color code is accepted')
        elif len(self.args) == 0:
            raise Exception('no color codes')
        return int(self.args[0])

    def __add__(self, other):
        if isinstance(other, Color):
            other = [int(i) for i in other.args]
        elif isinstance(other, list):
            other = [int(i) for i in other]
        elif isinstance(other, tuple):
            other = [int(i) for i in other]
        else:
            other = [int(other)]

        new_color = set(self.args + other)
                        
        return type(self)(*list(new_color))

    def __radd__(self, other):
        return self.__add__(other)

    #def __iadd__(self, other):
    #    return self.__add__(other)

    def __str__(self):
        return ';'.join( [str(i) for i in self.args] )

    def __repr__(self):
        return "<{}.{} object ({}) at {}>".format(
                    type(self).__module__,
                    type(self).__qualname__,
                    self.__str__(),
                    hex(id(self)))

    #def __iter__(self):
    #    for x in list.__iter__(self.args):
    #        yield x

    def __contains__(self, key):
        return int(key) in self.args

    
    def __getitem__(self, item):
        return self.args[item]


    def type(self):
        return type(self).type(self)

    @classmethod
    def type(cls, code):
        code = int(code)

        if code in cls._SET_:
            return 'SET'
        elif code in cls._RESET_:
            return 'RESET'
        elif code in cls._FDCOLOR_:
            return 'FDCOLOR'
        elif code in cls._FLCOLOR_:
            return 'FLCOLOR'
        elif code in cls._BDCOLOR_:
            return 'BDCOLOR'
        elif code in cls._BLCOLOR_:
            return 'BLCOLOR'
        elif code == 39:
            return 'FDEFAULT'
        elif code == 49:
            return 'BDEFAULT'

        raise Exception('Unknown color code')


    @classmethod
    def _sumList(cls, clist):
        if len(clist) == 0:
            return cls(0)

        c = None
        for t in clist:
            c = cls(t) if c is None else c+cls(t)
        return c

    @classmethod
    def reset(cls, *args):
        codes = []
        if len(args) == 0:
            return Color(0)
        else:

            for color in args:
                codes += list(color)

            for idx, code in enumerate(codes):
                t = cls.type(code)
                if t in ['SET']:
                    color = cls(code+20)
                elif t in ['FDCOLOR', 'FLCOLOR']:
                    color = cls(39)
                elif t in ['BDCOLOR', 'BLCOLOR']:
                    color = cls(49)
                else:
                    color = cls(code)

                codes[idx] = color
    
        return cls._sumList(codes)


    @classmethod
    def background(cls, *args):
        codes = []
        if len(args) == 0:
            raise Exception('no color was specified to be transformed into background color code')
        else:
            for color in args:
                codes += list(color)
            
            for idx, code in enumerate(codes):
                t = cls.type(code)
                if t in ['FDCOLOR', 'FLCOLOR', 'FDEFAULT']:
                    color = cls(code + 10)
                else:
                    color = cls(code)
                
                codes[idx] = color

        return cls._sumList(codes)
    
    @classmethod
    def foreground(cls, *args):
        codes = []
        if len(args) == 0:
            raise Exception('no color was specified to be transformed into forground color code')
        else:
            for color in args:
                codes += list(color)

            for idx, code in enumerate(codes):
                t = cls.type(code)
                if t in ['BDCOLOR', 'BLCOLOR', 'BDEFAULT']:
                    color = cls(code - 10)
                else:
                    color = cls(code)

                codes[idx] = color

        return cls._sumList(codes)


    @classmethod
    def dark(cls, *args):
        codes = []
        if len(args) == 0:
            raise Exception('no color was specified to be transformed into dark color code')
        else:
            for color in args:
                codes += list(color)
            
            for idx, code in enumerate(codes):
                t = cls.type(code)
                if t in ['BLCOLOR', 'FLCOLOR']:
                    color = cls(code - 60)
                else:
                    color = cls(code)

                codes[idx] = color

        return cls._sumList(codes)


    @classmethod
    def light(cls, *args):
        codes = []
        if len(args) == 0:
            raise Exception('no color was specified to be transformed into light color code')
        else:
            for color in args:
                codes += list(color)

            for idx, code in enumerate(codes):
                t = cls.type(code)
                if t in ['BDCOLOR', 'FDCOLOR']:
                    color = cls(code + 60)
                else:
                    color = cls(code)

                codes[idx] = color

        return cls._sumList(codes)


    @classmethod
    def bg(cls, *args):
        return cls.background(*args)

    @classmethod
    def fg(cls, *args):
        return cls.foreground(*args)





Color.BOLD = Color(1)
Color.DIM = Color(2)
Color.UNDERLINE = Color(4)
Color.BLINK = Color(5)
Color.REVERSE = Color(7)
Color.HIDDEN = Color(8)

Color.DEFAULT = Color(39)
Color.BLACK = Color(30)
Color.RED = Color(91)
Color.GREEN = Color(92)
Color.YELLOW = Color(93)
Color.BLUE = Color(94)
Color.MAGENTA = Color(95)
Color.CYAN = Color(96)
Color.GRAY = Color(37)
Color.WHITE = Color(97)




if __name__ == '__main__':

    print('Black:', Color.BLACK)
    print('White:', Color.WHITE)
    print('Bold:', Color.BOLD)
    print('Black + Bold:', Color.BOLD + Color.BLACK)
    print('reset all:', Color.reset())
    print('reset Black:', Color.reset(Color.BLACK))
    print('reset (Black + Bold):', Color.reset(Color.BLACK + Color.BOLD))
    print('reset (Black and Bold):', Color.reset(Color.BLACK, Color.BOLD))
    print('White background:', Color.bg(Color.WHITE))
    print('White background + Black foreground:', Color.bg(Color.WHITE) + Color.BLACK)
    print('reset background:', Color.reset(Color.bg(Color.WHITE)))
    print('reset background + reset foreground + bold:', Color.reset(Color.BLACK + Color.bg(Color.BLACK)) + Color.BOLD)



from .logger import getLogger
from .logger import LoggingConfig as Config

from .formatter import ColoredFormatter
from . import misc

from .logger import *


__all__ = [
    'getLogger',
    'Config',
    'ColoredFormatter',
    'misc'
]



# --- built in ---
import os
import sys
import json
import time
import smtplib

from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart


SECRET = None

def formatter(template, **kwargs):
    return template.format(**kwargs)

def setTemplate(template, formatter=formatter):
    '''
    Email template

    template: template file (*.html)
    formatter: a function to format template (default will use str.format)
    '''

    global SECRET

    if template is not None:
        with open(template, 'r') as f:
            SECRET['template'] = f.read()
            
    SECRET['formatter'] = formatter

def loadConfig(secret, template=None, formatter=formatter):
    '''
    Load gmail configurations from file

    sender: 'oooooo@gmail.com'
    password: 'xxxxxxxxxxx'
    smtp_ip: 'smtp.gmail.com' (Optional)
    smtp_port: 465 (Optional)
    targets: ['xxx001@gmail.com', 'xxx002@gmail.com'] (Optional)
    '''

    global SECRET
    with open(secret, 'r') as f:
        SECRET = json.load(f)

    assert SECRET.get('sender') and SECRET.get('password')

    SECRET['smtp_ip'] = SECRET.get('smtp_ip', 'smtp.gmail.com')
    SECRET['smtp_port'] = SECRET.get('smtp_port', 465)

    setTemplate(template, formatter)

    if SECRET.get('targets'):
        if isinstance(SECRET['targets'], str):
            SECRET['targets'] = [SECRET['targets']]
        assert isinstance(SECRET['targets'], list), 'targets must be a list of str'
    

def send(contents=None, 
         targets=None,
         subject=None, **kwargs):
    '''
    Send message
    '''

    global SECRET
    assert SECRET is not None, 'Call loadConfig first'

    if subject is None:
        subject = '[Blank]'

    # user information
    sender = SECRET['sender']
    password = SECRET['password']
    
    # email targets
    if targets is None:
        assert SECRET.get('targets'), 'Targets not specified'
    
        targets = SECRET['targets']

    targets = ', '.join(targets)

    # email template and formatter
    template = SECRET.get('template')
    formatter = SECRET.get('formatter')

    # create email
    mail = MIMEMultipart()
    mail['Subject'] = subject
    mail['From'] = sender
    mail['To'] = targets

    # format message
    if template:
        message = formatter(template, contents=contents,
                                      targets=targets,
                                      subject=subject,
                                      **kwargs)
    else:
        message = '{}'.format(contents)

    # attach message to email
    mail.attach(MIMEText(message, 'html'))

    # send email
    server = smtplib.SMTP_SSL(SECRET['smtp_ip'], SECRET['smtp_port'])
    server.login(SECRET['sender'], SECRET['password'])
    server.sendmail(sender, targets, mail.as_string())
    server.quit()